import io

from core.db_adapter.ceph.ceph_exception import CephIOMaxRetryReached, CephIOFileNotFoundException
from core.utils.rerunable import Rerunable


class CephIO(Rerunable):
    def __init__(self, config):
        super(CephIO, self).__init__(config)
        self.bucket = config["bucket"]
        self.filename = config["filename"]
        self.mode = config["mode"]

    @property
    def _handled_exception(self):
        return ConnectionError

    def _on_prepare(self):
        pass

    def _on_all_tries_fail(self):
        raise CephIOMaxRetryReached(f"Can't get key from {self.bucket} in case of {str(self._handled_exception)}.")

    def _get_bucket_keys(self):
        return self.bucket.get_key(self.filename)

    def __enter__(self):
        key = self._run(self._get_bucket_keys)
        if key:
            data = self._run(key.get_contents_as_string)
            io_stream = None
            if self.mode == "r":
                io_stream = io.StringIO(data.decode("utf-8"))
            elif self.mode == "rb":
                io_stream = io.BytesIO(data)
            return io_stream
        else:
            raise CephIOFileNotFoundException(f"{self.filename} is not exist.")

    def __exit__(self, type, value, traceback):
        pass



from typing import Dict

from core.configs.global_constants import CALLBACK_ID_HEADER, KAFKA_REPLY_TOPIC
from core.request.kafka_request import KafkaRequest


class SmartKitKafkaRequest(KafkaRequest):
    KAFKA_REPLY_TOPIC = KAFKA_REPLY_TOPIC
    KAFKA_EXTRA_HEADERS = "kafka_extraHeaders"

    def __init__(self, items, id=None):
        super(SmartKitKafkaRequest, self).__init__(items)
        items = items or {}
        self._callback_id = items.get(self._callback_id_header_name)
        self._kafka_replyTopic = items.get(self.KAFKA_REPLY_TOPIC)
        self._kafka_extraHeaders = items.get(self.KAFKA_EXTRA_HEADERS) or {}

    @property
    def _callback_id_header_name(self):
        return CALLBACK_ID_HEADER

    def _get_new_headers(self, source_mq_message):
        headers_dict = dict(super(SmartKitKafkaRequest, self)._get_new_headers(source_mq_message))
        if self._callback_id:
            headers_dict[self._callback_id_header_name] = str(self._callback_id).encode()
        if self._kafka_replyTopic:
            headers_dict[self.KAFKA_REPLY_TOPIC] = str(self._kafka_replyTopic).encode()
        if self._kafka_extraHeaders:
            for k, v in self._kafka_extraHeaders.items():
                headers_dict[k] = str(v).encode()
        headers_list = list(headers_dict.items())
        return headers_list

    def __str__(self):
        return f"KafkaRequest: kafka_key={self.kafka_key}"



import re

from core.text_preprocessing.grammem.grammem_constants import LEMMA, TEXT, TOKEN_TYPE, TOKEN_VALUE, VALUE, \
    LIST_OF_TOKEN_TYPES_DATA, ADJECTIVAL_NUMBER
from core.utils.utils import get_number

NUM_TOKEN = "NUM_TOKEN"

KILO_PREFIX = "к"
NUMBER_POS = 1
TOKEN_POS = 0
NOT_FROM_DICTIONARY_POS = 2
META_TOKEN_POS = 3
BEZ_POS = 4
ADJECTIVAL_POS = 5
NUMBER = "номер"
ADPOSITIONS = ["с", "от"]
BEZ_LIST = ["без", "Без"]
HOURS_RANGE = range(0, 21)
BEG_LIST = ["два", "три", "четыре"]


class Text2Num:
    """
    Класс предназначен для нахождения чисел в тексте и ковертации их в запись цифрами
    с приписыванием типа токена NUM_TOKEN
    """
    BLACK_LIST = set(range(10, 20))

    def __init__(self, text2num_dict_provider, black_list):
        self.text2num_dict_provider = text2num_dict_provider or {}
        self.black_list = set(black_list or [])
        self._regex_ending = re.compile("-[а-яё]{1,3}(\s|$)")
        self._multiplier_numbers = [1e1, 1e2, 1e3, 1e6, 1e9, 1e12, 1e15, 1e18, 12]
        self._ordinal_number_endings = {"ий", "ый"}

    def __convert(self, numbers_info):
        # агрегируем собранную информацию в N чисел <= M токенов
        result_numbers = [[numbers_info[0][TOKEN_POS], numbers_info[0][ADJECTIVAL_POS]]]
        if not result_numbers[0][0].isdigit():  # токен может быть как "100", так и "сто". В кейсах типа "сто",
            # т.е. когда изначально числительное представлено словом, мы берем цифру
            result_numbers = [[numbers_info[0][NUMBER_POS], numbers_info[0][ADJECTIVAL_POS]]]
        prev_number_info = numbers_info[0]
        for number_info in numbers_info[1:]:
            if number_info[NOT_FROM_DICTIONARY_POS]:
                result_numbers.append([number_info[TOKEN_POS], number_info[ADJECTIVAL_POS]])
            else:
                if number_info[NUMBER_POS] in self._multiplier_numbers and \
                        number_info[NUMBER_POS] >= prev_number_info[NUMBER_POS]:
                    result_numbers[-1] = [get_number(str(result_numbers[-1][0])) * number_info[NUMBER_POS],
                                          number_info[ADJECTIVAL_POS]]
                    # надо обработать без пятнадцати шесть
                elif prev_number_info[BEZ_POS] and number_info[NUMBER_POS] in HOURS_RANGE and not \
                        number_info[TOKEN_POS].isdigit() and (number_info[TOKEN_POS].endswith("ь") or \
                                                              number_info[TOKEN_POS].lower() in BEG_LIST):
                    result_numbers = [[number_info[NUMBER_POS], number_info[ADJECTIVAL_POS]] for number_info in
                                      numbers_info]
                elif number_info[TOKEN_POS].isdigit() and \
                        len(str(number_info[NUMBER_POS])) > len(str(prev_number_info[NUMBER_POS])):
                    # защита от кейса 6 5
                    result_numbers[-1] = [get_number(str(result_numbers[-1][0])) + number_info[NUMBER_POS],
                                          number_info[ADJECTIVAL_POS]]
                elif not number_info[TOKEN_POS].isdigit() and \
                        len(str(number_info[NUMBER_POS])) < len(str(prev_number_info[NUMBER_POS])) and not \
                        prev_number_info[NUMBER_POS] in self.BLACK_LIST:
                    # но надо аккуратно с "тысяча восемьсот пятьдесят три"
                    result_numbers[-1] = [get_number(str(result_numbers[-1][0])) + number_info[NUMBER_POS],
                                          number_info[ADJECTIVAL_POS]]
                else:
                    result_numbers = [[number_info[NUMBER_POS], number_info[ADJECTIVAL_POS]] for number_info in
                                      numbers_info]
                    break
            prev_number_info = number_info
        return result_numbers

    def _check_if_sberbank_phone(self, token, previous_token):
        is_sberbank_phone = False
        if token in self.black_list:
            if NUMBER in previous_token or previous_token in ADPOSITIONS:
                is_sberbank_phone = True
        return is_sberbank_phone

    def __get_number(self, token_desc, i, previous_token, token_desc_list):
        info_tuple = None
        token = token_desc[TEXT].lower()
        token = " ".join(self._regex_ending.sub(" ", token).split())
        previous_word_is_bez = False
        if token not in self._ordinal_number_endings:
            num = get_number(token)
            is_sberbank_phone = self._check_if_sberbank_phone(token, previous_token)
            if num is not None and not is_sberbank_phone:
                if i > 0 and previous_token in BEZ_LIST:
                    previous_word_is_bez = True
                info_tuple = (token, num, True, token_desc, previous_word_is_bez, False)
            else:
                if token == KILO_PREFIX:
                    if i > 0 and len(token_desc_list) > 1 and token_desc_list[i - 1][TEXT].isdigit():
                        num = self.text2num_dict_provider.get(token)
                else:
                    num = self.text2num_dict_provider.get(token)
                if num is not None:
                    if i > 0 and previous_token in BEZ_LIST:
                        previous_word_is_bez = True
                    value = num[0]
                    is_adjective = num[1]
                    info_tuple = (token, value, False, token_desc, previous_word_is_bez, is_adjective)
        return info_tuple

    def __merge(self, tuple_buffer_list, numbers_list):
        without_changes = len(tuple_buffer_list) == len(numbers_list)
        out_list = []
        if without_changes:
            for token_tuple, num in zip(tuple_buffer_list, numbers_list):
                number = num[0]
                is_adjective = num[1]
                token_desc = token_tuple[META_TOKEN_POS]
                token_desc[TEXT] = str(number)  # beware '1e-07', кладем такое в текст
                token_desc[LEMMA] = str(number)  # beware '1e-07', кладем такое в лемму
                token_desc[TOKEN_TYPE] = NUM_TOKEN
                token_desc[TOKEN_VALUE] = {VALUE: get_number(str(number)), ADJECTIVAL_NUMBER: is_adjective}
                preliminary_dicti = {}
                preliminary_dicti[TOKEN_TYPE] = NUM_TOKEN
                preliminary_dicti[TOKEN_VALUE] = {VALUE: get_number(str(number)), ADJECTIVAL_NUMBER: is_adjective}
                if LIST_OF_TOKEN_TYPES_DATA not in token_desc:
                    token_desc[LIST_OF_TOKEN_TYPES_DATA] = []
                token_desc[LIST_OF_TOKEN_TYPES_DATA].append(preliminary_dicti)
                out_list.append(token_desc)
        else:
            for num in numbers_list:
                number = num[0]
                is_adjective = num[1]
                t = {TEXT: str(number), LEMMA: str(number), TOKEN_TYPE: NUM_TOKEN,
                     TOKEN_VALUE: {VALUE: get_number(str(number)), ADJECTIVAL_NUMBER: is_adjective}}
                t[LIST_OF_TOKEN_TYPES_DATA] = [{TOKEN_TYPE: NUM_TOKEN,
                                                TOKEN_VALUE: {VALUE: get_number(str(number)),
                                                              ADJECTIVAL_NUMBER: is_adjective}}]
                # beware '1e-07' in text and lemma
                out_list.append(t)
        return out_list

    def __call__(self, token_desc_list, *args):
        """
        Метод находит все записи чисел в виде цифр (120.5) и букв (сто двадцать),
        конвертирует их при необходимости в циферную запись и проставляет тип токена NUM_TOKEN.
        :param token_desc_list: (list of dicts)
        :return: out: Список из dicts (enriched)
        """
        buffer = []
        out = []
        previous_token = ""
        for i, token in enumerate(token_desc_list):
            number_info_tuple = self.__get_number(token, i, previous_token, token_desc_list)
            if number_info_tuple is not None:
                buffer.append(number_info_tuple)
            elif len(buffer) > 0:
                numbers = self.__convert(buffer)
                token_list = self.__merge(buffer, numbers)
                out.extend(token_list)
                buffer = []
                out.append(token)
            else:
                out.append(token)
            previous_token = token[TEXT]

        if len(buffer) > 0:
            numbers = self.__convert(buffer)
            token_list = self.__merge(buffer, numbers)
            out.extend(token_list)
        return out


class NumbersUnionAfterSTT:
    """
    Объединение цифр в строке
    """
    BLACK_LIST = set(range(10, 20))

    def __init__(self, text2num_dict_provider):
        self.text2num_dict_provider = text2num_dict_provider or {}
        self._regex_ending = re.compile("-[а-яё]{1,3}(\s|$)")
        self._multiplier_numbers = [1e3, 1e6, 1e9, 1e12, 1e15, 1e18]
        self._ordinal_number_endings = {"ий", "ый"}

    def __convert(self, numbers_info):
        # агрегируем собранную информацию в N чисел <= M токенов
        result_numbers = [numbers_info[0][TOKEN_POS]]
        if not result_numbers[0].isdigit():  # токен может быть как "100", так и "сто". В кейсах типа "сто",
            # т.е. когда изначально числительное представлено словом, мы берем цифру
            result_numbers = [numbers_info[0][NUMBER_POS]]
        prev_number_info = numbers_info[0]
        for number_info in numbers_info[1:]:
            if number_info[NUMBER_POS] in self._multiplier_numbers and \
                    number_info[NUMBER_POS] >= prev_number_info[NUMBER_POS]:
                result_numbers[-1] = get_number(str(result_numbers[-1])) * number_info[NUMBER_POS]
            elif len(str(number_info[NUMBER_POS])) < len(str(prev_number_info[NUMBER_POS])) and not \
                    prev_number_info[NUMBER_POS] in self.BLACK_LIST and number_info[NUMBER_POS] != 0:
                result_numbers[-1] = get_number(str(result_numbers[-1])) + number_info[NUMBER_POS]
            else:
                result_numbers.append(number_info[NUMBER_POS])
            prev_number_info = number_info
        return result_numbers

    def __get_number(self, token_desc, i, token_desc_list):
        info_tuple = None
        token = token_desc.lower()
        token = " ".join(self._regex_ending.sub(" ", token).split())
        num = None
        if token not in self._ordinal_number_endings:
            if token == KILO_PREFIX:
                if i > 0 and len(token_desc_list) > 1 and token_desc_list[i - 1].isdigit():
                    num = self.text2num_dict_provider.get(token)[0]
            else:
                num = self.text2num_dict_provider.get(token)
            if num is not None:
                info_tuple = (token, num[0], False, token_desc)
        return info_tuple

    def __call__(self, input_string, *args):
        """
        Метод находит все записи чисел в виде цифр (120.5) и букв (сто двадцать),
        конвертирует их при необходимости в циферную запись и проставляет тип токена NUM_TOKEN.
        :param input_string: исходная текстовая строка
        :return: строка, где цифры уже объединены
        """
        buffer = []
        out = []
        token_desc_list = input_string.split(" ")
        for i, token in enumerate(token_desc_list):
            number_info_tuple = self.__get_number(token, i, token_desc_list)
            if number_info_tuple is not None:
                buffer.append(number_info_tuple)
            elif len(buffer) > 0:
                numbers = self.__convert(buffer)
                zero_counter = False
                token_list = []
                for k, number in enumerate(numbers):
                    if zero_counter:
                        token_list.append(str(0) + str(number))
                        zero_counter = False
                    elif number == 0 and k < len(numbers) - 1:
                        zero_counter = True
                    else:
                        token_list.append(str(number))
                out.extend(token_list)
                buffer = []
                out.append(token)
            else:
                out.append(token)

        if len(buffer) > 0:
            numbers = self.__convert(buffer)
            zero_counter = False
            token_list = []
            for k, number in enumerate(numbers):
                if zero_counter:
                    token_list.append(str(0) + str(number))
                    zero_counter = False
                elif number == 0 and k < len(numbers) - 1:
                    zero_counter = True
                else:
                    token_list.append(str(number))
            out.extend(token_list)
        return " ".join(out)



# coding: utf-8
import unittest

from core.configs import global_constants


class GlobalConstantsTest1(unittest.TestCase):
    def test_constants(self):
        self.assertIsNotNone(global_constants.CALLBACK_ID_HEADER)



from unittest import TestCase
from unittest.mock import Mock, MagicMock

from scenarios.user.last_scenarios.last_scenarios import LastScenarios
from smart_kit.utils.picklable_mock import PicklableMock


class ScenarioDescription:
    def __init__(self, dict):
        self._dict=dict

    def __contains__(self, item):
        return item in self._dict.keys()

    def __getitem__(self, item):
        return self._dict[item]


class TestLastScenarios(TestCase):

    def setUp(self):
        self.forms_dict = {'pay_phone_scenario':
                                {'fields': {'amount': {'value': 100.0}, 'approve': {'available': True, 'value': True}},
                                 'remove_time': 1506418333}, 'callcenter_scenario': {'remove_time': 2506421370}}

    def remove_item(self, item):
        self.forms_dict.pop(item)

    def test_add_1(self):
        items = ["pay_phone_scenario"]
        forms = MagicMock(name="context_form_mock")
        forms.clear_form = self.remove_item
        user = PicklableMock()
        user.forms = forms

        pay_phone_scenario = PicklableMock()
        pay_phone_scenario.keep_forms_alive = False
        pay_phone_scenario.form_type = "pay_phone_scenario"
        user.descriptions = {"scenarios": ScenarioDescription({"pay_phone_scenario": pay_phone_scenario})}
        last_scenario_description = PicklableMock()
        last_scenario_description.get_count = Mock(return_value=1)
        ls = LastScenarios(items, last_scenario_description, user)
        new_scenario = "callcenter_scenario"
        ls.add(new_scenario, None)
        self.assertSequenceEqual(["callcenter_scenario"], ls.raw)
        self.assertSequenceEqual({"callcenter_scenario": {"remove_time": 2506421370}}, self.forms_dict)

    def test_add_2(self):
        items = ["pay_phone_scenario"]
        forms = MagicMock(name="context_form_mock")
        user = PicklableMock()
        user.forms = forms

        pay_phone_scenario = PicklableMock()
        pay_phone_scenario.keep_forms_alive = False
        user.descriptions = {"scenarios": {"pay_phone_scenario": pay_phone_scenario}}
        last_scenario_description = PicklableMock()
        last_scenario_description.get_count = Mock(return_value=2)
        ls = LastScenarios(items, last_scenario_description, user)
        new_scenario = "callcenter_scenario"
        ls.add(new_scenario, None)
        self.assertSequenceEqual(["pay_phone_scenario", "callcenter_scenario"], ls.raw)
        self.assertSequenceEqual({"pay_phone_scenario":
                                      {"fields": {"amount": {"value": 100.0},
                                                  "approve": {"available": True, "value": True}},
                                       "remove_time": 1506418333},
                                  "callcenter_scenario": {"remove_time": 2506421370}}, self.forms_dict)

    def test_add_3(self):
        items = ["pay_phone_scenario"]
        forms = MagicMock(name="context_form_mock")
        user = PicklableMock()
        user.forms = forms

        pay_phone_scenario = PicklableMock()
        pay_phone_scenario.keep_forms_alive = True
        user.descriptions = {"scenarios": ScenarioDescription({"pay_phone_scenario": pay_phone_scenario})}
        last_scenario_description = PicklableMock()
        last_scenario_description.check = Mock(return_value=True)
        last_scenario_description.get_count = Mock(return_value=1)
        ls = LastScenarios(items, last_scenario_description, user)
        new_scenario = "callcenter_scenario"
        ls.add(new_scenario, None)
        correct_result = {'pay_phone_scenario': {'fields': {'amount': {'value': 100.0}, 'approve': {'available': True,
                                                                                                    'value': True}},
                          'remove_time': 1506418333}, 'callcenter_scenario': {'remove_time': 2506421370}}
        self.assertSequenceEqual(["callcenter_scenario"], ls.raw)
        self.assertSequenceEqual(correct_result, self.forms_dict)

    def test_add_4(self):
        items = ["pay_phone_scenario"]
        forms = MagicMock(name="context_form_mock")
        forms.clear_form = self.remove_item
        user = PicklableMock()
        user.forms = forms

        pay_phone_scenario = PicklableMock()
        pay_phone_scenario.keep_forms_alive = False
        pay_phone_scenario.form_type = "pay_phone_scenario"
        user.descriptions = {"scenarios": ScenarioDescription({"pay_phone_scenario": pay_phone_scenario})}
        last_scenario_description = PicklableMock()
        last_scenario_description.get_count = Mock(return_value=-1)
        ls = LastScenarios(items, last_scenario_description, user)
        new_scenario = "callcenter_scenario"
        ls.add(new_scenario, None)
        correct_result = {'callcenter_scenario': {'remove_time': 2506421370}}
        self.assertSequenceEqual(["callcenter_scenario"], ls.raw)
        self.assertSequenceEqual(correct_result, self.forms_dict)

    def test_clear_all(self):
        items = ["pay_phone_scenario"]
        forms = MagicMock(name="context_form_mock")
        user = PicklableMock()
        user.forms = forms

        pay_phone_scenario = PicklableMock()
        pay_phone_scenario.keep_forms_alive = False
        user.descriptions = {"scenarios": {"pay_phone_scenario": pay_phone_scenario}}

        last_scenario_description = PicklableMock()
        last_scenario_description.check = Mock(return_value=True)
        last_scenario_description.get_count = Mock(return_value=2)
        ls = LastScenarios(items, last_scenario_description, user)
        new_scenario = "callcenter_scenario"
        ls.add(new_scenario, None)
        ls.clear_all()
        self.assertSequenceEqual([], ls.raw)

    def test_get_last_scenario_1(self):
        items = ["pay_phone_scenario1", "pay_phone_scenario2"]
        user = PicklableMock()
        last_scenario_description = PicklableMock()
        ls = LastScenarios(items, last_scenario_description, user)
        self.assertEqual(ls.last_scenario_name, "pay_phone_scenario2")

    def test_get_last_scenario_2(self):
        items = []
        user = PicklableMock()
        last_scenario_description = PicklableMock()
        ls = LastScenarios(items, last_scenario_description, user)
        self.assertIsNone(ls.last_scenario_name)

    def test_expire(self):
        user = PicklableMock()
        user.forms = {"pay_phone_scenario": None}
        pay_phone_scenario = PicklableMock()
        pay_phone_scenario.form_type = "pay_phone_scenario"
        scenarios = ScenarioDescription({"pay_phone_scenario": pay_phone_scenario})
        user.descriptions = {"scenarios": scenarios}
        last_scenario_description = PicklableMock()
        items = ["pay_phone_scenario"]
        ls = LastScenarios(items, last_scenario_description, user)
        ls.expire()
        self.assertSequenceEqual([], ls.raw)


import csv
from core.repositories.base_repository import BaseRepository


class CSVRepository(BaseRepository):
    def __init__(self, filename, source=None, *args, **kwargs):
        super(CSVRepository, self).__init__(source=source, *args, **kwargs)
        self.filename = filename

    def load(self):
        with self.source.open(self.filename, newline='') as stream:
            reader = csv.DictReader(stream)
            data = list(reader)
            self.fill(data)
        super(CSVRepository, self).load()



# coding: utf-8
import unittest
from unittest.mock import Mock

from smart_kit.handlers import handler_base


class HandlersTest1(unittest.TestCase):
    def test_handler_base(self):
        self.test_user = Mock('user')
        self.test_user.message = Mock("messsage")
        self.test_user.message.message_name = "test"
        self.test_user.message.channel = "test_channel"
        self.test_user.message.device = Mock("device")
        self.test_user.message.device.surface = "test_surface"
        obj = handler_base.HandlerBase("TestAppName")
        self.assertIsNotNone(obj.TOPIC_KEY)
        self.assertIsNotNone(obj.KAFKA_KEY)



# coding: utf-8
import unittest

from scenarios.scenario_models.field_requirements.field_requirements import TokenPartInSet


class RequirementTest(unittest.TestCase):

    def test_token_part_in_set_requirement_equal_false(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": "locality_type",
              "values": ["DISTRICT", "REGION"]
            }
        token_val = {'value': 'Амадора',
                     'locality_type': 'CITY',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_true(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": "locality_type",
              "values": ["DISTRICT", "CITY"]
            }
        token_val = {'value': 'Амадора',
                     'locality_type': 'CITY',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertTrue(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_False_double_empty(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": "locality_type",
              "values": []
            }
        token_val = {'value': 'Амадора',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False,
                     'locality_type': []}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_False_empty_val_none(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": "locality_type",
              "values": []
            }
        token_val = {'value': 'Амадора',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False,
                     'locality_type': None}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_False_string(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": "value",
              "values": 'cba'
            }
        token_val = {'value': 'abc',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_False(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": "country_hidden",
              "values": [1, 2, 3]
            }
        token_val = {'value': 'Амадора',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_False_val_int(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": 'capital',
              "values": [-9.23083]
            }
        token_val = {'value': 'Амадора',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_True_arr(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": 'timezone',
              "values": [[[None, 1.0]]]
            }
        token_val = {'value': 'Амадора',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertTrue(requirement.check(token_val))

    def test_token_part_in_set_requirement_equal_False_arr(self):
        requirement_items ={
              "type": "token_part_in_set",
              "part": 'timezone',
              "values": [[[1.0, None]]]
            }
        token_val = {'value': 'Амадора',
                     'latitude': 38.75382,
                     'longitude': -9.23083,
                     'capital': None,
                     'locative_value': None,
                     'timezone': [[None, 1.0]],
                     'currency': ['EUR', 'евро'],
                     'country': 'Португалия',
                     'country_hidden': False}
        requirement = TokenPartInSet(requirement_items)
        self.assertFalse(requirement.check(token_val))



import copy

import aioredis
import typing

from core.db_adapter.db_adapter import DBAdapter
from core.db_adapter import error
from core.monitoring.monitoring import monitoring

from core.logging.logger_utils import log


class AIORedisAdapter(DBAdapter):
    IS_ASYNC = True

    def __init__(self, config=None):
        super().__init__(config)
        self._redis: typing.Optional[aioredis.Redis] = None

        try:
            del self.config["type"]
        except KeyError:
            pass

    @monitoring.got_histogram("save_time")
    async def save(self, id, data):
        return await self._run(self._save, id, data)

    @monitoring.got_histogram("save_time")
    async def replace_if_equals(self, id, sample, data):
        return await self._run(self._replace_if_equals, id, sample, data)

    @monitoring.got_histogram("get_time")
    async def get(self, id):
        return await self._run(self._get, id)

    async def path_exists(self, path):
        return await self._run(self._path_exists, path)

    async def connect(self):
        print("Here is the content of REDIS_CONFIG:", self.config)
        print("Connecting to a single redis server")
        config = copy.deepcopy(self.config)
        redis_url = config.pop("redis", None)
        if not isinstance(redis_url, str):
            raise ValueError(
                "redis should be specified like redis://[[username]:[password]]@localhost:6379/0")
        self._redis = aioredis.from_url(redis_url, **config)

    def _open(self, filename, *args, **kwargs):
        pass

    async def _save(self, id, data):
        return await self._redis.set(id, data)

    async def _replace_if_equals(self, id, sample, data):
        log("no replace_if_equals method. running save instead", level="WARNING")
        await self.save(id, data)

    async def _get(self, id):
        data = await self._redis.get(id)
        return data

    def _list_dir(self, path):
        raise error.NotSupportedOperation

    def _glob(self, path, pattern):
        raise error.NotSupportedOperation

    async def _path_exists(self, path):
        return await self._redis.exists(path)

    def _on_prepare(self):
        pass



# coding: utf-8
from scenarios.scenario_models.field.field import field_model_factory
from scenarios.scenario_models.field.fields import Fields


class CompositeField:
    def __init__(self, description, items, user, lifetime):
        self.description = description
        self._user = user
        self._lifetime = lifetime
        self.fields = Fields(items.get("fields"), description.fields, user, field_model_factory, lifetime)

    def fill(self, value):
        any_filled = False
        if value:
            for field_descr in self.fields:
                filled = self.fields[field_descr].fill(value.get(field_descr.id))
                any_filled = any_filled or filled
        # when any inner field is filled at this run and all fields are valid for the time
        return any_filled and self.valid

    @property
    def value(self):
        value = {}
        for field_descr in self.fields:
            field = self.fields[field_descr]
            if field.value:
                value[field_descr.id] = field.value
        return value

    @property
    def available(self):
        return all([self.fields[field_descr].available for field_descr in self.fields])

    def set_available(self):
        for field_descr in self.fields:
            self.fields[field_descr].set_available()

    @property
    def valid(self):
        return all(self.fields[field_descr].valid for field_descr in self.fields)

    @property
    def raw(self):
        fields = {}
        for field_descr in self.fields:
            field = self.fields[field_descr]
            if field.raw:
                fields[field_descr.id] = field.raw
        raw = {}
        if fields:
            raw["fields"] = fields
        return raw



from functools import cached_property
from typing import List, Sequence

import nltk
from rusenttokenize import ru_sent_tokenize

from core.repositories.file_repository import FileRepository
from core.utils.loader import ordered_json

from smart_kit.text_preprocessing.base_text_normalizer import BaseTextNormalizer
from smart_kit.text_preprocessing.nltk_tokenizer_binding import NLTKWordTokenizer
from smart_kit.text_preprocessing.pymorphy2_morph_wrapper import Pymorphy2MorphWrapper
from smart_kit.text_preprocessing.text2num import Text2Num, NumbersUnionAfterSTT
from smart_kit.text_preprocessing.utils import unmerge_numbers_and_letters, merge_numbers, \
    replace_currencies_symbols, AdditionalMathSplitter, MergeCardNumbersVoice, MergeCardNumbers, \
    NormalizePhoneNumbersVoice, NormalizePhoneNumbers, UnicodeSymbolsConverter, ReplaceSynonyms, \
    CurrencyTokensOneIterationMerger, reverse_json_dict, return_lemmas_only, Singleton


class LocalTextNormalizer(BaseTextNormalizer, metaclass=Singleton):
    """
    Упрощённая предобработка, соответствующая первым шагам платформенной
    Разбиение на слова, замена числительных и валют, морфология. Извлечение сущностей отсутствует
    При вызове нужно передавать параметр message_type="voice" или ="text" - для голоса и текста разный пайплайн
    """

    def __init__(self):
        self.__ready_to_use = False
        self._morph = None

    @cached_property
    def morph(self):
        self._morph = Pymorphy2MorphWrapper()
        return self._morph

    def get_token_list(self, text):
        sentences = self.sentence_tokenizer(text.strip())
        words = self.word_tokenizer(sentences)
        return words

    def __load_everything(self):
        try:
            nltk.data.find("tokenizers/punkt")
        except LookupError:
            nltk.download("punkt")

        from smart_kit.configs import get_app_config
        app_config = get_app_config()
        text_normalizer_params = FileRepository(
            f"{app_config.STATIC_PATH}/.text_normalizer_resources/static_workdata.json",
            loader=ordered_json,
        )
        text_normalizer_params.load()

        synonyms = FileRepository(
            f"{app_config.STATIC_PATH}/.text_normalizer_resources/dict_synonyms.json",
            loader=reverse_json_dict,
        )
        synonyms.load()
        text_normalizer_params.data["synonyms"] = synonyms.data

        text2num_dict = FileRepository(
            f"{app_config.STATIC_PATH}/.text_normalizer_resources/text2num_dict.json",
            loader=ordered_json,
        )
        text2num_dict.load()
        text_normalizer_params.data["text2num_dict"] = text2num_dict.data

        text_normalizer_params = text_normalizer_params.data or {}
        self.convert_plan = text_normalizer_params["convert_plan"]
        self.processor_pipeline = text_normalizer_params["processor_pipeline"]

        word_false_stoppings = text_normalizer_params.get("word_false_stoppings", [])
        words_without_splitting_point = text_normalizer_params.get("word_no_splitting_point", [])
        synonyms = text_normalizer_params.get("synonyms", {})
        text2num = text_normalizer_params.get("text2num_dict", {})
        sberbank_phones = text_normalizer_params.get("sberbank_phones", [])
        unicode_symbols = text_normalizer_params.get("unicode_symbols", {})

        self.sentence_tokenizer = ru_sent_tokenize
        self.word_tokenizer = NLTKWordTokenizer(word_false_stoppings, words_without_splitting_point)

        skip_func = lambda x: x
        self.converter_pipeline = {
            'Объединение цифр после stt': NumbersUnionAfterSTT(text2num),
            'Конверсия юникодовых символов': UnicodeSymbolsConverter(unicode_symbols),
            'Цифры и буквы отдельно': unmerge_numbers_and_letters,
            'Номера телефонов': NormalizePhoneNumbers(),
            'Номера телефонов из голоса': NormalizePhoneNumbersVoice(),
            'Номера карт': MergeCardNumbers(),
            'Номера карт из голоса': MergeCardNumbersVoice(),
            'Объединение сумм': merge_numbers,
            'Символы валют': replace_currencies_symbols,
            "Претокенизация математических операций": AdditionalMathSplitter()
        }
        self.tokens_processor = {
            "Synonyms": ReplaceSynonyms(synonyms) if synonyms else skip_func,
            "Text2Num": Text2Num(text2num, sberbank_phones) if text2num else skip_func,
            "Currency": CurrencyTokensOneIterationMerger(),
            "Grammemes": self.morph,
        }

        self.__ready_to_use = True

    def load_everything(self):
        if not self.__ready_to_use:
            self.__load_everything()

    def __call__(self, text, message_type="voice"):
        self.load_everything()
        normalized_text = text
        convert_processor = self.convert_plan[message_type]
        for converter_name in convert_processor:
            converter = self.converter_pipeline[converter_name]
            normalized_text = converter(normalized_text)
        token_list = self.get_token_list(text)
        for converter_step in self.processor_pipeline:
            converter_name = converter_step["name"]
            converter = self.tokens_processor[converter_name]
            token_list = converter(token_list)
        return {
            "original_text": text,
            "normalized_text": return_lemmas_only(token_list),
            "tokenized_elements_list": token_list
        }

    @classmethod
    def with_cache(cls, app_config, **kwargs):
        return cls()

    def normalize_sequence(self, texts: Sequence, batch_size=None) -> List:
        normalized_texts = []
        for text in texts:
            normalized_texts.append(self(text))
        return normalized_texts



from unittest import TestCase

from core.unified_template.unified_template import UnifiedTemplate, UNIFIED_TEMPLATE_TYPE_NAME


class TestUnifiedTemplate(TestCase):
    def test_ordinar_jinja_backward_compatibility(self):
        template = UnifiedTemplate("abc {{input}}")
        self.assertEqual("abc def", template.render({"input": "def"}))

    def test_type_cast_int(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "loader": "int",
            "template": "{{ input ** 2 }}"
        })
        self.assertEqual(49, template.render({"input": 7}))

    def test_type_cast_json(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "loader": "json",
            "template": "{{ input }}"
        })
        self.assertEqual([1, 2, 3], template.render({"input": [1, 2, 3]}))

    def test_support_templates(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "support_templates": {
                "name": "{{ personInfo.name|capitalize }}",
                "surname": "{{ personInfo.surname|capitalize }}"
            },
            "template": "{{ name }} {{ surname }}",
        })
        self.assertEqual("Vasya Pupkin", template.render({"personInfo": {"name": "vasya", "surname": "pupkin"}}))

    def test_args_format_input(self):
        template = UnifiedTemplate("timestamp: {{ts}}")
        self.assertEqual("timestamp: 123.45", template.render(ts=123.45))

    def test_bool_cast_false(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "loader": "bool",
            "template": "False"
        })
        self.assertFalse(template.render({}))

    def test_bool_cast_false_lowercase(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "loader": "bool",
            "template": "false"
        })
        self.assertFalse(template.render({}))

    def test_bool_cast_true(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "loader": "bool",
            "template": "True"
        })
        self.assertTrue(template.render({}))

    def test_bool_cast_true_lowercase(self):
        template = UnifiedTemplate({
            "type": UNIFIED_TEMPLATE_TYPE_NAME,
            "loader": "bool",
            "template": "true"
        })
        self.assertTrue(template.render({}))



# coding: utf-8
from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems
from scenarios.scenario_models.forms.form_description import form_description_factory


class FormsDescription(SmartUpdatableDescriptionsItems):
    def __init__(self, items):
        super(FormsDescription, self).__init__(form_description_factory, items)



import re
import pymorphy2
from core.text_preprocessing.grammem.grammem_constants import GRAMMEM_INFO, PART_OF_SPEECH, LEMMA, TEXT, TOKEN_TYPE, \
    LIST_OF_TOKEN_TYPES_DATA, TOKEN_VALUE, VALUE, RAW_GRAM_INFO, OTHER, TRANSITIVITY, ANIMACY, ASPECT

SENTENCE_ENDPOINT_TOKEN = "SENTENCE_ENDPOINT_TOKEN"


def token_list_to_sentences(token_desc_list: list) -> list:
    """
    Разбить список токенов на предложения
    :param token_desc_list: список токенов
    :return: Список из списков, где первый список содержит предложения, а второй токены
    """
    results = []
    k = 0
    more_than_one_sentence = False
    for i, token in enumerate(token_desc_list):
        if token.get(TOKEN_TYPE) == SENTENCE_ENDPOINT_TOKEN:
            more_than_one_sentence = True
            results.append(token_desc_list[k:i])
            k = i + 1
    if not more_than_one_sentence:
        results = [token_desc_list]
    return results


class Pymorphy2MorphWrapper:
    """
    Класс предназначен для получения граммемной информации о токенах.
    """

    def __init__(self):
        self.pymorphy_analyzer = pymorphy2.MorphAnalyzer()
        self.latin = re.compile("^[0-9]*[A-Za-z]+[0-9]*$")
        self.cyrillic = re.compile("[А-Яа-яЁе]+")

    def _choose_pymorphy_form(self, word, lemma, pos):
        hypotheses = self.pymorphy_analyzer.parse(word)
        hyp = None
        tags_to_add = {}
        other = ""
        for hyp in hypotheses:
            if hyp.normal_form == lemma:
                break
        changed_lemma = lemma
        if not hyp:
            return other, tags_to_add, changed_lemma
        str_tag = str(hyp.tag)
        if "Surn" in str_tag:
            other = "фам"
            changed_lemma = word.lower()
        elif "Patr" in str_tag:
            other = "отч"
            changed_lemma = word.lower()
        if hyp.tag.transitivity:
            tags_to_add[TRANSITIVITY] = str(hyp.tag.transitivity)
        if hyp.tag.animacy and pos == "NOUN":
            tags_to_add[ANIMACY] = str(hyp.tag.animacy)
        if hyp.tag.aspect:
            tags_to_add[ASPECT] = str(hyp.tag.aspect)
        return other, tags_to_add, changed_lemma

    def _change_pos(self, token, analysis):
        if re.match(self.latin, analysis.word):
            token[GRAMMEM_INFO][PART_OF_SPEECH] = "X"
        elif analysis.tag.cyr_repr == "ЗПР" and re.search(self.cyrillic, analysis.word):
            token[GRAMMEM_INFO][PART_OF_SPEECH] = "X"
        else:
            token[GRAMMEM_INFO][PART_OF_SPEECH] = analysis.tag.POS
        return token

    def _gram_info_processing(self, tags_to_add, analysis):
        gramme_info = {}
        raw_gram_data = []
        tags = ["_POS", "animacy", "aspect", "case", "gender", "involvement", "mood", "number", "person", "tense",
                "transitivity", "voice"]
        for key in tags:
            if getattr(analysis.tag, key):
                gramme_info[key] = getattr(analysis.tag, key)
        gramme_info.update(tags_to_add)
        sorted_gramme_info = {key: gramme_info[key] for key in sorted(gramme_info.keys())}
        for key in sorted_gramme_info:
            raw_gram_data.append(key + "=" + sorted_gramme_info[key])
        raw_gram_info = "|".join(raw_gram_data)
        return sorted_gramme_info, raw_gram_info

    def _pymorphy_morph_to_token_dicti(self, token, analysis):
        additional_info, tags_to_add, changed_lemma = self._choose_pymorphy_form(analysis.word, analysis.normal_form,
                                                                                 analysis.tag.POS)
        sorted_gramme_info, raw_gram_info = self._gram_info_processing(tags_to_add, analysis)
        token[GRAMMEM_INFO] = sorted_gramme_info
        token[GRAMMEM_INFO][RAW_GRAM_INFO] = raw_gram_info
        if additional_info:
            token[GRAMMEM_INFO][OTHER] = additional_info
        token = self._change_pos(token, analysis)
        token[LEMMA] = changed_lemma
        return token

    def token_desc_list_processing(self, token_desc_list):
        """
        Получить список токенов с описанием
        :param: Список из словарей
        :return: Список из словарей, обогащенный морфологической информацией
        """
        raw_token_list = [token[TEXT] for token in token_desc_list]
        analyze_result = [self.pymorphy_analyzer.parse(word)[0] for word in raw_token_list]

        res = []
        for i in range(len(token_desc_list)):
            analysis = analyze_result[i]
            tokenized_element = token_desc_list[i]
            final_tokenized_element = self._pymorphy_morph_to_token_dicti(tokenized_element, analysis)
            res.append(final_tokenized_element)
        return res

    def __call__(self, token_desc_list):
        """
        Класс предназначен для забора из RNNMorph + pymorphy2 граммемной информации.
        На вход принимается список токенов
        На выходе имеем список токенов с проставленными грамматическими атрибутами

        :param token_desc_list (list of dicts)
        :return: final_result (enriched list of dicts)
        """
        final_result = []
        sentences = token_list_to_sentences(token_desc_list)
        for sentence in sentences:
            final_result.extend(self.token_desc_list_processing(sentence))
            if final_result:
                final_result.append({TEXT: ".", LEMMA: ".", TOKEN_TYPE: SENTENCE_ENDPOINT_TOKEN,
                                     TOKEN_VALUE: {VALUE: "."},
                                     LIST_OF_TOKEN_TYPES_DATA:
                                         [{TOKEN_TYPE: SENTENCE_ENDPOINT_TOKEN, TOKEN_VALUE: {VALUE: "."}}]
                                     })
        return final_result



# coding: utf-8
import unittest

from core.db_adapter import memory_adapter


class AdapterTest1(unittest.TestCase):

    def test_memory_adapter_init(self):
        obj1 = memory_adapter.MemoryAdapter()
        obj2 = memory_adapter.MemoryAdapter({'try_count': 3})
        self.assertTrue(hasattr(obj1, 'open'))
        self.assertTrue(hasattr(obj2, 'open'))
        self.assertTrue(obj1.memory_storage == {})
        self.assertTrue(obj2.memory_storage == {})
        self.assertTrue(obj1.try_count == 5)  # взято из исходников
        self.assertTrue(obj2.try_count == 3)

    def test_memory_adapter_connect(self):
        obj = memory_adapter.MemoryAdapter()
        self.assertTrue(hasattr(obj, 'connect'))

    def test_memory_adapter_open(self):
        obj = memory_adapter.MemoryAdapter()
        self.assertTrue(hasattr(obj, '_open'))
        with self.assertRaises(TypeError):
            obj._open()

    def test_memory_adapter_save(self):
        obj1 = memory_adapter.MemoryAdapter()
        obj2 = memory_adapter.MemoryAdapter()
        obj1._save(10, {'any_data'})
        obj1._save(11, {'any_data'})
        obj2._save(10, {'any_data'})
        self.assertTrue(obj1.memory_storage == {10: {'any_data'}, 11: {'any_data'}})
        self.assertTrue(obj2.memory_storage == {10: {'any_data'}})
        # метод переписывает значения
        obj2._save(10, 'any_data')
        self.assertTrue(obj2.memory_storage == {10: 'any_data'})

    def test_memory_adapter_get(self):
        obj1 = memory_adapter.MemoryAdapter()
        obj1._save(10, {'any_data'})
        obj1._save(11, 'any_data')
        self.assertTrue(obj1._get(10) == {'any_data'})
        self.assertTrue(obj1._get(11) == 'any_data')
        self.assertIsNone(obj1._get(12))

    def test_memory_adapter_list_dir(self):
        obj = memory_adapter.MemoryAdapter()
        self.assertTrue(hasattr(obj, '_list_dir'))
        with self.assertRaises(TypeError):
            obj._open()



from unittest import TestCase

from scenarios.scenario_models.field.field_filler_description import AllRegexpsFieldFiller
from smart_kit.utils.picklable_mock import PicklableMock


class Test_regexps_filler(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.items = {}
        cls.items["exps"] = ["номер[а-я]*\.?\s?(\d+)", "n\.?\s?(\d+)", "nn\.?\s?(\d+)", "#\.?\s?(\d+)",
                             "##\.?\s?(\d+)",
                             "№\.?\s?(\d+)", "№№\.?\s?(\d+)", "платеж[а-я]+\.?\s?(\d+)", "поручен[а-я]+\.?\s?(\d+)",
                             "п\\s?,\\s?п\.?\s?(\d+)", "п\\s?\\/\\s?п\.?\s?(\d+)"]
        cls.items["delimiter"] = "|"

        cls.filler = AllRegexpsFieldFiller(cls.items)

    def test_extract_1(self):
        field_value = "Просим отозвать платежное поручение 14 от 23.01.19 на сумму 3500 и вернуть деньги на расчетный счет."
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = AllRegexpsFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, None)
        self.assertEqual('14', result)

    def test_extract_2(self):
        field_value = "поручение12 поручение14 #1 n3 п/п70 n33"
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = AllRegexpsFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, None)
        self.assertEqual("3|33|1|12|14|70", result)

    def test_extract_no_match(self):
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = "текст без искомых номеров"

        filler = AllRegexpsFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)



from functools import cached_property

from core.model.factory import dict_factory, factory
from scenarios.scenario_models.field.field_filler_description import FieldFillerDescription
from scenarios.scenario_models.field.field_descriptions.basic_field_description import BasicFieldDescription


class CompositeFieldDescription(BasicFieldDescription):

    def __init__(self, items, id):
        super(CompositeFieldDescription, self).__init__(items, id)
        self._fields = items["fields"]

    @cached_property
    @factory(FieldFillerDescription)
    def filler(self):
        return self._filler

    @cached_property
    @dict_factory(BasicFieldDescription)
    def fields(self):
        return self._fields



import logging.config
from collections import namedtuple

from smart_kit.configs.logger_config import LOGGING_CONFIG
from smart_kit.configs.logger_config import LoggerConfig
from smart_kit.management.base import HelpCommand
from smart_kit.management.cache import CreateCacheCommand
from smart_kit.management.tests import TestsCommand
from smart_kit.management.plugins import activate_plugins
from smart_kit.start_points.app import run as app_runner
from smart_kit.management.get_bundles_from_pps import GetBundleCommand


def init_logger(app_config):
    logger_config = LoggerConfig(app_config.CONFIGS_PATH)
    logger_config.init()
    config = logger_config[LOGGING_CONFIG]
    logging.log_store_for_map = config.get('log_store_for_map', None)
    logging.config.dictConfig(config)


class AppCommand:
    @classmethod
    def doc(cls):
        return cls.__doc__ or "No help provided"

    def execute(self, *args, **kwargs):
        raise NotImplementedError


class LocalTestingCommand(AppCommand):
    def __init__(self, app_config):
        self.app_config = app_config
        self._local_testing_wrapper = app_config.LOCAL_TESTING(
            configs_path=app_config.CONFIGS_PATH, secret_path=app_config.SECRET_PATH,
            settings_cls=app_config.SETTINGS, references_path=app_config.REFERENCES_PATH,
            resources_cls=app_config.RESOURCES, model_cls=app_config.MODEL, test_case_cls=app_config.TEST_CASE,
            dialogue_manager_cls=app_config.DIALOGUE_MANAGER, user_cls=app_config.USER,
            parametrizer_cls=app_config.PARAMETRIZER, from_msg_cls=app_config.FROM_MSG, app_name=app_config.APP_NAME
        )

    def execute(self, *args, **kwargs):
        init_logger(self.app_config)
        if self.app_config.NORMALIZER:
            self.app_config.NORMALIZER.load_everything()
        return self._local_testing_wrapper.run_local_testing(*args, **kwargs)


class RunAppCommand(AppCommand):
    def __init__(self, app_config):
        self.app_config = app_config
        self._runner = app_runner

    def execute(self, *args, **kwargs):
        init_logger(self.app_config)
        if self.app_config.NORMALIZER:
            self.app_config.NORMALIZER.load_everything()
        return self._runner(self.app_config)


class AppManager:
    def __init__(self):
        self.commands = {}
        self.Command = namedtuple('Command', 'command_class args kwargs')

    def register_command(self, name, command, *args, **kwargs):
        self.commands[name] = self.Command(command, args, kwargs)

    def execute_command(self, name, *args, **kwargs):
        current_command = self.commands[name]
        command = current_command.command_class(*current_command.args, **current_command.kwargs)
        return command.execute(*args, **kwargs)

    def execute_from_command_line(self, argv):
        if len(argv) == 1:
            name = "help"
        else:
            name = argv[1]

        return self.execute_command(name, *argv[2:])


def execute_from_command_line(argv):
    from smart_kit.configs import get_app_config
    app_config = get_app_config()
    manager = AppManager()
    manager.register_command("local_testing", LocalTestingCommand, app_config)
    manager.register_command("run_app", RunAppCommand, app_config)
    manager.register_command("tests", TestsCommand, app_config)
    manager.register_command("cache", CreateCacheCommand, app_config)
    manager.register_command("help", HelpCommand, manager.commands)
    manager.register_command("get_bundles", GetBundleCommand, app_config)

    activate_plugins(app_config, manager)

    return manager.execute_from_command_line(argv)



# coding: utf-8
import json
from functools import cached_property
from typing import List

from core.descriptions.descriptions import Descriptions
from core.model.queued_objects.limited_queued_hashable_objects_description import \
    LimitedQueuedHashableObjectsDescription
from core.model.queued_objects.limited_queued_hashable_objects import LimitedQueuedHashableObjects, \
    LimitedQueuedHashableObjectsItems
from core.logging.logger_utils import log
from core.model.field import Field
from core.model.model import Model
from core.basic_models.parametrizers.parametrizer import BasicParametrizer
from core.basic_models.counter.counters import Counters
from core.basic_models.variables.variables import Variables
from scenarios.logging.logger_constants import KEY_NAME


class BaseUser(Model):
    USER_DB_VERSION = "user_db_data_version"

    counters: Counters
    variables: Variables
    private_vars: Variables
    local_vars: Variables
    descriptions: Descriptions

    def __init__(self, id, message, values, descriptions, load_error=False):
        self.id = id
        self.message = message
        self.descriptions = descriptions
        self.load_error = load_error
        super().__init__(values=values, user=self)
        self._initialize()

    def _initialize(self):
        """Add here some things to do after creating fields."""
        pass

    @cached_property
    def parametrizer(self):
        return BasicParametrizer(self, {})

    @property
    def fields(self) -> List[Field]:
        return [
            Field("counters", Counters),
            Field("last_action_ids", LimitedQueuedHashableObjectsItems, self.descriptions["last_action_ids"]),
            Field("last_messages_ids", LimitedQueuedHashableObjects, LimitedQueuedHashableObjectsDescription(None)),
            Field("local_vars", Variables, None, False),
            Field("variables", Variables),
            Field("private_vars", Variables),
        ]

    @property
    def raw_str(self):
        # Attention: non-serializable objects will become str with error message
        raw = json.dumps(self.raw, default=lambda o: f"<non-serializable: {type(o).__qualname__}>")
        log("%(class_name)s.raw USER %(uid)s SAVE db_version = %(db_version)s. "
            "Saving User %(uid)s. Serialized utf8 json length is %(user_length)s symbols.", self,
            {"db_version": str(self.private_vars.get(self.USER_DB_VERSION)),
             "uid": str(self.id), "user_length": len(raw),
             KEY_NAME: "user_save"})
        return raw

    def expire(self):
        self.counters.expire()
        self.variables.expire()
        self.private_vars.expire()
        self.local_vars.expire()



from core.text_preprocessing.grammem.grammem_constants import (
    IS_STOP_WORD, LEMMA, PART_OF_SPEECH, GRAMMEM_INFO, TEXT, TOKEN_TYPE, LIST_OF_DEPENDENTS, AMBIGUOUS, TENSE, MOOD,
    COMPOSITE_TOKEN_TYPE, IS_BEGINNING_OF_COMPOSITE, COMPOSITE_TOKEN_LENGTH,
    W2V_STRING, DEPENDENT_ADPOSITIONS, IMPORTANT_SURNAME_VALUE,
    IS_BEGINNING_OF_ANAPHOR, ANTECEDENT_LEMMA, PART_OF_SPEECH_PYMORPHY)
from core.text_preprocessing.constants import SENTENCE_ENDPOINT_TOKEN, ANIMACY_TOKEN


class TokenizeHelper:
    @staticmethod
    def lemma_modification(lemma: str) -> str:
        if len(lemma) > 1 and "." in lemma:
            res = lemma.strip(".")
            if "." in res:
                res = res.replace(".", "_")
        else:
            res = lemma
        return res

    @staticmethod
    def token_list_to_sentences(token_desc_list: list) -> list:
        """
        Разбить список токенов на предложения
        :param token_desc_list: список токенов
        :return: Список из списков, где первый список содержит предложения, а второй токены
        """
        results = []
        k = 0
        more_than_one_sentence = False
        for i, token in enumerate(token_desc_list):
            if token.get(TOKEN_TYPE) == SENTENCE_ENDPOINT_TOKEN:
                more_than_one_sentence = True
                results.append(token_desc_list[k:i])
                k = i + 1
        if not more_than_one_sentence:
            results = [token_desc_list]
        return results

    @staticmethod
    def get_tense(token):
        result = ''
        if token.get(GRAMMEM_INFO, {}).get(PART_OF_SPEECH) == 'V':
            if AMBIGUOUS in token[GRAMMEM_INFO]:
                result = '{}_{}'.format(result, token[GRAMMEM_INFO][AMBIGUOUS][0].get(TENSE, ''))
            else:
                result = '{}_{}'.format(result, token[GRAMMEM_INFO].get(TENSE, ''))
        return result

    @staticmethod
    def get_mood(token):
        result = ''
        if token.get(GRAMMEM_INFO, {}).get(PART_OF_SPEECH) == 'V':
            if AMBIGUOUS in token[GRAMMEM_INFO]:
                result = '{}_{}'.format(result, token[GRAMMEM_INFO][AMBIGUOUS][0].get(MOOD, ''))
            else:
                result = '{}_{}'.format(result, token[GRAMMEM_INFO].get(MOOD, ''))

        return result

    @staticmethod
    def return_lemmas_only(token_desc_list: list, include_sentence_endpoint: bool = True, show_verb_mood: bool =
    False, consider_stop_words: bool = False) -> str:
        """
        Конвертировать список токенов в строку без пунктуации, содержащую леммы и замены на тип токена.
        Предложения разделяются точкой. Если перед этим нет грамматической инфо - возвращается поле токена TEXT
        :param token_desc_list: Список токенов (дикты), include_sentence_endpoint -- принтить ли фейковую точку, bool
        :return: Строка, содержащую леммы и замены на тип токена.
        """
        final_line = []
        for token in token_desc_list:
            if COMPOSITE_TOKEN_TYPE in token:
                if token[IS_BEGINNING_OF_COMPOSITE]:
                    if token.get(IMPORTANT_SURNAME_VALUE):
                        token_type = token[IMPORTANT_SURNAME_VALUE]
                    else:
                        token_type = token[COMPOSITE_TOKEN_TYPE]
                else:
                    continue
            else:
                token_type = token.get(TOKEN_TYPE)
            if token_type != SENTENCE_ENDPOINT_TOKEN:
                grammem_info = token.get(GRAMMEM_INFO)
                if grammem_info is not None:
                    pos = grammem_info.get(PART_OF_SPEECH_PYMORPHY)
                    if pos != "PNCT":
                        if token_type is not None and token_type != ANIMACY_TOKEN:
                            final_line.append(token_type)
                        elif not token.get(IS_STOP_WORD) or (token.get(IS_STOP_WORD) and consider_stop_words):
                            lemma = TokenizeHelper.lemma_modification(token.get(LEMMA))
                            if show_verb_mood:
                                lemma = '{}{}{}'.format(lemma, TokenizeHelper.get_tense(token),
                                                        TokenizeHelper.get_mood(token))
                            final_line.append(lemma)
                else:
                    final_line.append(token[TEXT])
            elif include_sentence_endpoint:
                final_line.append(token[TEXT])
        return " ".join(final_line)

    @staticmethod
    def return_lemmas_only_from_TNR(text_normalization_result) -> str:
        """
        Конвертировать список токенов в строку без пунктуации, содержащую леммы и замены на тип токена.
        Предложения разделяются точкой
        :param text_normalization_result: TextNormalizationResult
        :return: Строка, содержащую леммы и замены на тип токена.
        """
        tokens = text_normalization_result.tokenized_elements_list
        return TokenizeHelper.return_lemmas_only(tokens)

    @staticmethod
    def num_tokens(text_normalization_result) -> int:
        """
         Вернуть количество токенов в textnormalizationresult'е за вычетом фейковых точек
         Стоп-слова при этом считаются, решили, что правильно их считать
         :param text_normalization_ result: TextNormalizationResult
         :return: Int -- количество токенов
         """
        tokens_number = 0
        for token in text_normalization_result.tokenized_elements_list:
            if COMPOSITE_TOKEN_TYPE in token:
                if token[IS_BEGINNING_OF_COMPOSITE]:
                    tokens_number += 1
            elif token.get(TOKEN_TYPE) != SENTENCE_ENDPOINT_TOKEN:
                tokens_number += 1
        return tokens_number

    @staticmethod
    def get_adpositions_from_dependents(list_of_tokens: list, token: dict, list_of_adpositions: list) -> list:
        """
         Вернуть все предлоги, зависящие от данного токеноа
         :param list_of_tokens: list, token: dict, list_of_adpositions: list
         :return: list_of_adpositions -- массив зависимых предлогов
         """
        list_of_dependent_adpositions = []
        for dependent_token_idx in token.get(LIST_OF_DEPENDENTS, []):
            dependent_token = list_of_tokens[dependent_token_idx - 1]
            dependent_lemma = dependent_token.get(LEMMA, '')
            dependent_is_part_of_hypertoken = dependent_token.get(COMPOSITE_TOKEN_TYPE, '')
            if dependent_lemma in list_of_adpositions and not dependent_is_part_of_hypertoken:
                list_of_dependent_adpositions.append(dependent_lemma)
        return list_of_dependent_adpositions

    @staticmethod
    def tokens_to_w2v(list_of_tokens: list, list_of_adpositions: list, include_sentence_endpoint: bool = True) -> list:
        """
         Вернуть упрощенный список токенов, только w2v-представление и наличие предлога
         для глаголов добавляется время и наклонение(глагол_время_наклонение)
         :param list_of_tokens: list
         :return: list_of_tokens: list
         """
        short_list = []
        composite_beginning_index = 0
        composite_dependent_adpositions = False
        for i, token in enumerate(list_of_tokens):
            string_for_w2v = ""
            dependent_adpositions = TokenizeHelper.get_adpositions_from_dependents(list_of_tokens, token,
                                                                                   list_of_adpositions)
            if COMPOSITE_TOKEN_TYPE in token:
                if token[IS_BEGINNING_OF_COMPOSITE]:
                    composite_beginning_index = i
                    has_length_one = token[COMPOSITE_TOKEN_LENGTH] == 1
                    final_dependent_adpositions = []  ##надо не потерять у гипертокенов
                    final_dependent_adpositions.extend(dependent_adpositions)
                    if not has_length_one:  ##не проскипать кейс, когда этот же токен и есть последний
                        continue
                if i - composite_beginning_index != token[COMPOSITE_TOKEN_LENGTH] - 1:  ##кладем последний токен,
                    ##когда есть только первый компонент гипертокена, мы еще не знаем всех зависимых предлогов
                    final_dependent_adpositions.extend(dependent_adpositions)
                    continue
                else:
                    final_dependent_adpositions.extend(dependent_adpositions)
                    composite_dependent_adpositions = True
                    if token.get(IMPORTANT_SURNAME_VALUE):
                        token_type = token[IMPORTANT_SURNAME_VALUE]
                    else:
                        token_type = token[COMPOSITE_TOKEN_TYPE]
            else:
                token_type = token.get(TOKEN_TYPE)
            if token_type != SENTENCE_ENDPOINT_TOKEN:
                grammem_info = token.get(GRAMMEM_INFO, {})
                pos = grammem_info.get(PART_OF_SPEECH)
                if pos != "PUNCT":
                    if token_type is not None and token_type != ANIMACY_TOKEN:
                        string_for_w2v = token_type
                    elif not token.get(IS_STOP_WORD):
                        lemma = TokenizeHelper.lemma_modification(token.get(LEMMA))
                        lemma = '{}{}{}'.format(lemma, TokenizeHelper.get_tense(token), TokenizeHelper.get_mood(token))
                        string_for_w2v = lemma
            elif include_sentence_endpoint:
                string_for_w2v = token[TEXT]
            if composite_dependent_adpositions:
                composite_dependent_adpositions = False
                final_token = {W2V_STRING: string_for_w2v, DEPENDENT_ADPOSITIONS: final_dependent_adpositions}
                short_list.append(final_token)
            elif string_for_w2v:
                final_token = {W2V_STRING: string_for_w2v, DEPENDENT_ADPOSITIONS: dependent_adpositions}
                short_list.append(final_token)
        return short_list

    @staticmethod
    def return_indices_of_the_dependents(list_of_tokens: list, index: int) -> list:
        """
         Вернуть индексы зависимых от токена, где токен представлен индексом
         :param list_of_tokens: list, index: int
         :return: list_of_dependent_indices: list of ints
         """
        converted_elem = list_of_tokens[index]
        preliminary_list = converted_elem.get(LIST_OF_DEPENDENTS, [])
        list_of_dependent_indices = [index - 1 for index in preliminary_list]
        return list_of_dependent_indices

    @staticmethod
    def get_human_normalized_text(list_of_tokens: list):
        return " ".join([t[LEMMA] for t in list_of_tokens if t.get(TOKEN_TYPE) != SENTENCE_ENDPOINT_TOKEN
                         and LEMMA in t])

    @staticmethod
    def get_human_normalized_text_with_anaphora(list_of_tokens: list):
        list_of_lemmas = []
        for t in list_of_tokens:
            if t.get(IS_BEGINNING_OF_ANAPHOR):
                list_of_lemmas.append(t[ANTECEDENT_LEMMA])
            elif t.get(IS_BEGINNING_OF_ANAPHOR) == False:
                continue
            elif t.get(TOKEN_TYPE) != SENTENCE_ENDPOINT_TOKEN:
                list_of_lemmas.append(t[LEMMA])
        return " ".join(list_of_lemmas)



from scenarios.scenario_models.forms.form import BaseForm
from scenarios.scenario_models.forms.forms import Forms


class CompositeForm(BaseForm):
    def __init__(self, items, description, user):
        super(CompositeForm, self).__init__(items, description, user)
        items = items or {}
        self.forms = Forms(items.get("forms"), description.forms, user)
        self._valid = items.get("valid") or self.description.valid

    def is_valid(self):
        return self._valid

    def set_valid(self):
        self._valid = True

    def get_fields_values(self):
        data = {}
        for form in self.description.forms.keys():
            inner_form = self.forms[form]
            fields = inner_form.get_fields_values() if inner_form else {form: {}}
            data.update(fields)
        return {self.description.id: data}

    @property
    def raw(self):
        raw = {"valid": self._valid}
        if self.forms.raw:
            raw["forms"] = self.forms.raw
        if self.remove_time:
            raw["remove_time"] = self.remove_time
        return raw



# coding: utf-8
import time
from functools import cached_property

from core.basic_models.actions.basic_actions import Action
from core.model.factory import factory


class BehaviorDescription:
    def __init__(self, items, id=None):
        self.id = id
        self._success_action = items["success_action"]
        self._fail_action = items.get("fail_action")
        self._misstate = items.get("misstate")
        self._timeout_action = items.get("timeout_action")
        self._timeout = items.get("timeout", 300)
        self.version = items.get("version", -1)
        self.loop_def = items.get("loop_def", True)

    def get_expire_time_from_now(self, user):
        return time.time() + self.timeout(user)

    def timeout(self, user):
        setting_timeout = user.settings["template_settings"].get("services_timeout", {}).get(self.id)
        return setting_timeout or self._timeout

    @cached_property
    @factory(Action)
    def success_action(self):
        return self._success_action

    @cached_property
    @factory(Action)
    def fail_action(self):
        return self._fail_action

    @cached_property
    @factory(Action)
    def misstate(self):
        return self._misstate

    @cached_property
    @factory(Action)
    def timeout_action(self):
        return self._timeout_action



import functools
import re
import typing
import tabulate
import json

DoesNotExpected = object()


class Diff:
    def __init__(self):
        self.missed: typing.Dict[str, typing.Any] = {}  # key : expected
        self.different: typing.Dict[str, typing.Tuple[typing.Any, typing.Any]] = {}  # key: expected,actual
        self.does_not_expected: typing.Dict[str, typing.Any] = {}  # key: actual

    def update(self, other):
        self.missed.update(other.missed)
        self.different.update(other.different)
        self.does_not_expected.update(other.does_not_expected)

    def __bool__(self):
        return bool(self.missed or self.different)

    def __str__(self):
        missed = ""
        does_not_expected = ""
        different = ""
        if self.missed:
            missed = "Missed:\n" + tabulate.tabulate(self.missed.items(), headers=["Key", "Expected"])
        if self.does_not_expected:
            does_not_expected = "Does Not Expected:\n"
            does_not_expected += tabulate.tabulate(self.does_not_expected.items(), headers=["Key", "Actual"])
        if self.different:
            different = "Different:\n"
            different += tabulate.tabulate(
                [(k, d[0], d[1]) for k, d in self.different.items()], headers=["Key", "Expected", "Actual"]
            )

        return "\n".join((missed, does_not_expected, different))

    def serialize(self):
        return json.dumps({
            'missed': self.missed.copy(),
            'different': self.different.copy()
        })


@functools.singledispatch
def partial_diff(expected: typing.Any, actual: typing.Any, path="") -> Diff:
    diff = Diff()
    if expected != actual and \
            not (isinstance(actual, str) and isinstance(expected, str) and re.fullmatch(expected, actual)):
        diff.different[path] = expected, actual
    return diff


@partial_diff.register(list)
def partial_diff_list(expected: list, actual: list, path="") -> Diff:
    diff = Diff()
    does_not_expected = [DoesNotExpected for _ in range(len(expected), len(actual))]

    for index, exp_element in enumerate(expected + does_not_expected):
        sub_path = f"{path}[{index + 1}]" if path else f"[{index + 1}]"
        try:
            if exp_element == DoesNotExpected:
                diff.does_not_expected[sub_path] = actual[index]
            else:
                diff.update(partial_diff(exp_element, actual[index], sub_path))
        except IndexError:
            diff.missed[sub_path] = exp_element

    return diff


@partial_diff.register(dict)
def partial_diff_dict(expected, actual, path="") -> Diff:
    diff = Diff()
    for key, value in expected.items():
        sub_path = f"{path}.{key}" if path else str(key)
        if key not in actual:
            diff.missed[sub_path] = value
            continue

        diff.update(partial_diff(value, actual[key], sub_path))
    return diff



import os
from logging.handlers import RotatingFileHandler


class RotatingFilePidHandler(RotatingFileHandler):

    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0, encoding=None, delay=False):
        pid = os.getpid()
        filename = f"{filename}.{pid}"
        super(RotatingFilePidHandler, self).__init__(filename, mode, maxBytes, backupCount, encoding, delay)



from unittest import TestCase

from core.basic_models.actions.basic_actions import Action, action_factory, actions, DoingNothingAction, RandomAction
from core.model.registered import registered_factories


class TestRandomAction(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        registered_factories[Action] = action_factory
        actions["do_nothing"] = DoingNothingAction

    def test_1(self):

        items = {
            "actions": [
                {
                    "type": "do_nothing",
                    "command": "ANSWER_TO_USER",
                    "nodes": {
                        "answer": "Доброе утро!",
                    }
                },
                {
                    "type": "do_nothing",
                    "command": "ANSWER_TO_USER",
                    "nodes": {
                        "answer": "Добрый вечер!",
                    }
                }
            ]
        }
        action = RandomAction(items, 5)
        result = action.run(None, None)
        self.assertIsNotNone(result)

    def test_2(self):
        items = {
            "actions": [
                {
                    "type": "do_nothing",
                    "command": "ANSWER_TO_USER",
                    "nodes": {
                        "answer": "Добрый вечер!",
                    }
                }
            ]
        }
        action = RandomAction(items, 5)
        result = action.run(None, None)
        self.assertIsNotNone(result)



# coding: utf-8
import random

import pyignite
from pyignite.exceptions import ReconnectError, SocketError

import core.logging.logger_constants as log_const
from core.db_adapter import error
from core.db_adapter.db_adapter import DBAdapter
from core.logging.logger_utils import log
from core.monitoring.monitoring import monitoring


class IgniteAdapter(DBAdapter):

    def __init__(self, config):
        self._init_params = config.get("init_params", {})
        self._url = config["url"]
        if config.get("randomize_url"):
            random.shuffle(self._url)
        self._cache_name = config["cache_name"]
        self._cache = None
        super(IgniteAdapter, self).__init__(config)

    def _open(self, filename, *args, **kwargs):
        pass

    def _list_dir(self, path):
        raise error.NotSupportedOperation

    def _glob(self, path, pattern):
        raise error.NotSupportedOperation

    def _path_exists(self, path):
        raise error.NotSupportedOperation

    def connect(self):
        try:
            self._client = pyignite.Client(**self._init_params)
            self._client.connect(self._url)
            self._cache = self._client.get_or_create_cache(self._cache_name)
            logger_args = {
                log_const.KEY_NAME: log_const.IGNITE_VALUE,
                "pyignite_addresses": str(self._url)
            }
            log("IgniteAdapter to servers %(pyignite_addresses)s created", params=logger_args, level="WARNING")
        except Exception:
            log("IgniteAdapter connect error",
                          params={log_const.KEY_NAME: log_const.HANDLED_EXCEPTION_VALUE},
                          level="ERROR",
                          exc_info=True)
            monitoring.got_counter("ignite_connection_exception")
            raise

    def _save(self, id, data):
        return self.cache.put(id, data)

    def _replace_if_equals(self, id, sample, data):
        return self._cache.replace_if_equals(id, sample, data)

    def _get(self, id):
        data = self.cache.get(id)
        return data

    @property
    def cache(self):
        if self._cache is None:
            log('Attempt to recreate ignite instance', level="WARNING")
            self.connect()
            monitoring.got_counter("ignite_reconnection")
        return self._cache

    @property
    def _handled_exception(self):
        # TypeError is raised during reconnection if all nodes are exhausted
        return OSError, SocketError, ReconnectError

    def _on_prepare(self):
        self._cache = None

    def _get_counter_name(self):
        return "ignite_adapter"



import time
from unittest import TestCase

from scenarios.user.last_fields.last_fields import LastFields


class LastFieldsTest(TestCase):
    def test_expire_some(self):
        current_time = time.time()
        field_1 = {
            "value": 1,
            "remove_time": current_time + 100
        }
        field_2 = {
            "value": 2,
            "remove_time": current_time - 1
        }
        name_1 = "test_filed_1"
        fields = LastFields({name_1: field_1, "test_field_2": field_2}, None)
        fields.expire()
        self.assertDictEqual(fields.raw, {name_1: field_1})

    def test_expire_all(self):
        remove_time = time.time() - 100
        field_1 = {
            "value": 1,
            "remove_time": remove_time
        }

        field_2 = {
            "value": 2,
            "remove_time": remove_time - 100
        }
        fields = LastFields({"test_field_1": field_1, "test_field_2": field_2}, None)
        fields.expire()
        self.assertDictEqual(fields.raw, {})

    def test_expire_none(self):
        current_time = time.time() + 100
        field_1 = {
            "value": 1,
            "remove_time": current_time
        }
        field_2 = {
            "value": 2,
            "remove_time": current_time + 100
        }
        name_1 = "test_field_1"
        name_2 = "test_field_2"
        fields = LastFields({name_1: field_1, name_2: field_2}, None)
        fields.expire()
        self.assertDictEqual(fields.raw, {name_1: field_1, name_2: field_2})

    def test_expire_not_set(self):
        field_1 = {"value": 1}
        field_2 = {"value": 2}
        name_1 = "test_field_1"
        name_2 = "test_field_2"
        fields = LastFields({name_1: field_1, name_2: field_2}, None)
        fields.expire()
        self.assertDictEqual(fields.raw, {name_1: field_1, name_2: field_2})



import fnmatch
import ssl

import boto
import boto.s3.connection as s3_connection
from boto.exception import BotoServerError, BotoClientError

import core.logging.logger_constants as log_const
from core.db_adapter.ceph.ceph_io import CephIO
from core.db_adapter.db_adapter import DBAdapter
from core.logging.logger_utils import log
from core.monitoring.monitoring import monitoring

ssl._create_default_https_context = ssl._create_unverified_context


class CephAdapter(DBAdapter):
    def __init__(self, config):
        self._bucket = None
        super(CephAdapter, self).__init__(config)

    def connect(self):
        try:
            self._client = boto.connect_s3(
                aws_access_key_id=self.config["access_key"],
                aws_secret_access_key=self.config["secret_key"],
                host=self.config["host"],
                is_secure=self.config["is_secure"],
                port=self.config["port"],
                calling_format=s3_connection.OrdinaryCallingFormat(),
            )
            self._bucket = self._client.get_bucket(self.config["bucket"])
        except Exception:
            log("CephAdapter connect error",
                params={log_const.KEY_NAME: log_const.HANDLED_EXCEPTION_VALUE},
                level="ERROR",
                exc_info=True)
            monitoring.got_counter("ceph_connection_exception")
            raise

    @property
    def _handled_exception(self):
        return BotoServerError, BotoClientError

    @property
    def source(self):
        return self

    def _list_dir(self, path):
        return [key.name for key in self._bucket.list(prefix=path)]

    def _open(self, filename, mode, *args, **kwargs):
        io_config = {"filename": filename,
                     "mode": mode,
                     "bucket": self._bucket,
                     "try_count": self.config.get("read_tries")}
        return CephIO(io_config)

    def _on_prepare(self):
        self.connect()

    def _get_counter_name(self):
        return "ceph_adapter"

    def _glob(self, path, pattern):
        files_list = self._list_dir(path)
        filtered = fnmatch.filter(files_list, pattern)
        return filtered

    def _path_exists(self, path):
        return bool(self._list_dir(path))

    def _mtime(self, path):
        key = self._bucket.get_key(path)
        if key and key.last_modified:
            return key.last_modified



from typing import Dict, Any, Optional, Union

from core.basic_models.requirement.basic_requirements import Requirement
from core.model.factory import build_factory, factory
from core.model.registered import Registered

answer_items = Registered()
items_factory = build_factory(answer_items)

ANSWER_TO_USER = "ANSWER_TO_USER"


class SdkAnswerItem:
    version: int
    id: Optional[str]
    requirement: Requirement

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        items = items or {}
        self.id = id
        self.version = items.get("version", -1)
        self._requirement = items.get("requirement", None)
        self.requirement = self.build_requirement()

    def render(self, nodes: Dict[str, Any]):
        return {}

    @factory(Requirement)
    def build_requirement(self) -> str:
        return self._requirement


class TextSdkItem(SdkAnswerItem):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(TextSdkItem, self).__init__(items, id)
        self.text = items["text"]


class BubbleText(TextSdkItem):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(BubbleText, self).__init__(items, id)
        self.markdown = items.get("markdown", True)

    def render(self, nodes: Dict[str, Any]) -> Dict[str, Dict[str, str]]:
        return {
            "bubble": {
                "text": nodes.get(self.text, self.text),
                "markdown": self.markdown
            }
        }


class ItemCard(TextSdkItem):
    def render(self, nodes: Dict[str, Any]) -> Dict[str, str]:
        return {"card": nodes.get(self.text, self.text)}


class PronounceText(TextSdkItem):
    def render(self, nodes: Dict[str, Any]) -> Dict[str, str]:
        return {"pronounceText": nodes.get(self.text, self.text)}


class SuggestText(TextSdkItem):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(SuggestText, self).__init__(items, id)
        self.title = items["title"]

    def render(self, nodes: Dict[str, Any]) -> Dict[str, Union[str, Dict[str, str]]]:
        return {
            "title": nodes.get(self.title, self.title),
            "action": {
                "text": nodes.get(self.text, self.text),
                "type": "text"
            }
        }


class SuggestDeepLink(SdkAnswerItem):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(SuggestDeepLink, self).__init__(items, id)
        self.title = items["title"]
        self.deep_link = items["deep_link"]

    def render(self, nodes: Dict[str, Any]) -> Dict[str, Union[str, Dict[str, str]]]:
        return {
            "title": nodes.get(self.title, self.title),
            "action": {
                "deep_link": nodes.get(self.deep_link, self.deep_link),
                "type": "deep_link"
            }
        }


class RawItem(SdkAnswerItem):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(RawItem, self).__init__(items, id)
        self.key = items["key"]
        self.value = items["value"]

    def render(self, nodes: Dict[str, Any]) -> Dict[str, str]:
        return {self.key: nodes.get(self.value, self.value)}



from typing import Dict, Any, Optional

from core.text_preprocessing.base import BaseTextPreprocessingResult

from core.basic_models.requirement.basic_requirements import Requirement
from scenarios.user.user_model import User


class SettingsRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)
        self._config = items.get("config", "template_settings")
        self._key = items["key"]
        self._value = items["value"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        return user.settings[self._config][self._key] == self._value



import re
from nltk import word_tokenize
from core.text_preprocessing.grammem.grammem_constants import TOKEN_TYPE, TEXT
from .utils import DateConverter

SENTENCE_ENDPOINT_TOKEN = "SENTENCE_ENDPOINT_TOKEN"


class NLTKWordTokenizer:
    URL = re.compile(
        "(?:(?:https?|ftp)://)?(?:www)?[a-zA-Zа-яА-Я0-9_-]+(?:\\.[a-zA-Zа-яА-Я0-9_-]+)?\\.(?:com|ru|рф|org|bz|ua|ge|by|mp|net|cn)(?:/[a-zA-Zа-яА-Я0-9_\\-\\#\\&\\=\\?\\(\\)]+)*(?:\\.html|\\.xml)?/?")
    MAIL = re.compile(r"[\w\.\+\-]+\@[\w]+\.[a-z]{2,3}")
    SPACES = re.compile("^\s+$")
    QUASIPUNCT = "=…№-——'"
    DATE_CHECKER = DateConverter()

    def __init__(self, stoppings_list, black_list):
        self.false_stoppings = stoppings_list
        self.black_list = black_list

    def url_mail_fix(self, text):
        urls = re.findall(NLTKWordTokenizer.URL, text)
        mails = re.findall(NLTKWordTokenizer.MAIL, text)
        hard_tokens = mails + urls
        return hard_tokens

    def _non_last_removal(self, cnt, tokens, bad_token, token):
        to_return = 0
        length = len(tokens)
        if cnt == length - 1:
            to_return = 2
        elif cnt < length - 1:
            next_token = tokens[cnt + 1]
            if next_token in bad_token and next_token not in bad_token[:len(token)]:
                to_return = 1
            else:
                to_return = 0
        return to_return

    def _tokens_resolution(self, list_of_tokens, list_of_urls_mails, dicti, final_tokens, flag_to_append):
        final_token = ""
        for i, token in enumerate(list_of_tokens):
            flag_not_change = False
            if not flag_to_append:
                for hard in list_of_urls_mails:
                    if token == hard[:len(token)] and dicti[hard] == 0:
                        dicti[hard] = 1
                        flag_to_append = True
                        flag_not_change = True
                        final_token = token
                        if i == len(list_of_tokens) - 1:
                            final_tokens.append(token)
                        break
                if not flag_not_change:
                    final_tokens.append(token)
            elif flag_to_append:
                if token in hard and token not in hard[:len(token)]:  # для отсечки дополнительных https://
                    final_token += token
                    returned = self._non_last_removal(i, list_of_tokens, hard, token)
                    if returned == 2:
                        final_tokens.append(final_token)
                    elif returned == 1:
                        continue
                    else:
                        flag_to_append = False
                        final_tokens.append(final_token)
                else:
                    flag_to_append = False
                    final_tokens.append(final_token)
        return final_tokens

    def web_postprocessing(self, text, all_tokens):
        url_mails = self.url_mail_fix(text)
        if len(url_mails) > 0:
            dicti = {key: 0 for key in url_mails}
            final_tokens = self._tokens_resolution(all_tokens, url_mails, dicti, [], False)
            return final_tokens
        return all_tokens

    def endwise_removal(self, token):
        i = len(token) - 1
        buffer = []
        while token[i] in NLTKWordTokenizer.QUASIPUNCT:
            buffer.insert(0, token[i])
            i -= 1
        return buffer

    def _quasipunct_removal(self, token, new_token, future_tokens):
        if len(new_token) != 0:
            i = 0
            while token[i] in NLTKWordTokenizer.QUASIPUNCT:
                future_tokens.append({TEXT: token[i]})
                i += 1
            else:
                future_tokens.append({TEXT: new_token})
                buffer = self.endwise_removal(token)
                for elem in buffer:
                    future_tokens.append({TEXT: elem})
        else:
            if token == "''" or token == '``':
                future_tokens.append({TEXT: '"'})
            else:
                future_tokens.append({TEXT: token})
        return future_tokens

    def _slash_splitting(self, token):
        is_url = re.search(NLTKWordTokenizer.URL, token)
        is_date = NLTKWordTokenizer.DATE_CHECKER(token) is not None
        if "/" in token and not is_date and not is_url:
            tokens = []
            split_tokens = token.split("/")
            for i, elem in enumerate(split_tokens):
                if elem:
                    tokens.append({TEXT: elem})
                    if i != len(split_tokens) - 1:
                        tokens.append({TEXT: "/"})
        elif "\\" in token and not is_date and not is_url:
            tokens = [{TEXT: elem} for elem in token.split("\\") if elem]
        else:
            tokens = [{TEXT: token}]
        return tokens

    def punct_postprocessing(self, arr_of_tokens):
        final_tokens = []
        skip = False
        for i, token in enumerate(arr_of_tokens):
            new_token = token.strip(NLTKWordTokenizer.QUASIPUNCT)
            if skip:
                skip = False
            elif len(new_token) != len(token):
                final_tokens = self._quasipunct_removal(token, new_token, final_tokens)
            elif len(token) == 0:
                continue
            elif re.match("\s+", token):
                continue
            elif "/" in token or "\\" in token:
                if token == "/":
                    final_tokens.append({TEXT: "/"})
                else:
                    final_tokens.extend(self._slash_splitting(token))
            elif token == '``':
                final_tokens.append({TEXT: '"'})
            elif token == "''":
                final_tokens.append({TEXT: '"'})
            elif token == "..":
                final_tokens.append({TEXT: "."})
                final_tokens.append({TEXT: "."})
            elif len(arr_of_tokens) > i + 1 and token in self.false_stoppings and arr_of_tokens[
                i + 1] == ".":
                final_tokens.append({TEXT: token + "."})
                skip = True
            elif "." in token and len(token) == 3 and len(arr_of_tokens) > i + 1 and arr_of_tokens[i + 1] == ".":
                final_tokens.append({TEXT: token + "."})
                skip = True
            else:
                final_tokens.append({TEXT: token})
        return final_tokens

    def _preprocess_check(self, token):
        final_list = []
        if not re.match(NLTKWordTokenizer.SPACES, token):
            token = token.strip()
        if "." in token:
            if re.search(NLTKWordTokenizer.MAIL, token):
                final_list.append(token + " ")
            elif re.search(NLTKWordTokenizer.URL, token):
                final_list.append(token + " ")
            elif re.match(NLTKWordTokenizer.DATE_CHECKER.regex, token):
                match_result = re.match(NLTKWordTokenizer.DATE_CHECKER.regex, token)
                result = match_result.group(0)
                finish = match_result.end()
                final_list.append(result + " ")
                final_list.append(token[finish:] + " ")
            else:
                mask_token = token.lower().replace(".", "")
                if mask_token.isdigit():
                    final_list.append(token + " ")
                elif mask_token in self.black_list:
                    final_list.append(token + " ")
                else:
                    new_tokens = token.split(".")
                    for new_token in new_tokens[:-1]:
                        new_string = new_token + ". "
                        final_list.append(new_string)
                    final_list.append(new_tokens[-1] + " ")
        else:
            final_list.append(token + " ")
        return "".join(final_list)

    def preprocess(self, sentence):
        final_list = []
        if " " in sentence:
            for token in sentence.split(" "):
                final_list.append(self._preprocess_check(token))
        else:
            final_list.append(self._preprocess_check(sentence))
        return "".join(final_list).strip()

    def _merge_hashtags(self, tokens):
        '''
        слепляем токен '#' с идущим за ним
        :param tokens: list of strings
        :return: list of strings
        '''
        merged_tokens = []
        i = 0
        while i < len(tokens):
            if tokens[i] == '#' and i + 1 < len(tokens):
                merged_tokens.append(tokens[i] + tokens[i + 1])
                i += 2
            else:
                merged_tokens.append(tokens[i])
                i += 1

        return merged_tokens

    def __call__(self, list_of_texts):
        """
        Класс предназначен для токенизации (т.е. разбиения на "слова" предложений).
        На вход принимается список предложений (строк)
        На выходе имеем список словарей, в котором в том месте, где была граница между предложениями,
        стоит специальная синтетическая точка с типом токена SENTENCE_ENDPOINT_TOKEN

        :param list_of_texts (list of strings)
        :return: final_list (list of dicts)
        """
        final_list = []
        for text in list_of_texts:
            text = self.preprocess(text)
            preliminary_tokens = word_tokenize(text)
            preliminary_tokens = self._merge_hashtags(preliminary_tokens)
            final_arr = self.web_postprocessing(text, preliminary_tokens)
            post_arr = self.punct_postprocessing(final_arr)
            final_list.extend(post_arr)
            final_list.append({TEXT: ".", TOKEN_TYPE: SENTENCE_ENDPOINT_TOKEN})
        return final_list



import datetime
from unittest import TestCase

import jinja2

import core.unified_template.jinja_filters
from core.unified_template.currency2text import Money2Text
from core.unified_template.num2ordinal import Num2Ordinal
from core.unified_template.num2text import Num2Text


class TestJinjaTemplates(TestCase):
    def setUp(self):
        self.num2text = Num2Text()
        self.num2ord = Num2Ordinal()
        self.money2text = Money2Text()

    def test_inflect(self):
        template = jinja2.Template(
            "Поставим слово '{{ word }}' в множественное число, родительный падеж: '{{ word|inflect(['plur', 'gent']) }}'")
        self.assertEqual(
            "Поставим слово 'бутявка' в множественное число, родительный падеж: 'бутявок'",
            template.render({"word": "бутявка"})
        )

    def test_inflect_several_words_with_register(self):
        template = jinja2.Template("{{text|inflect(['loct'])}}")
        self.assertEqual(
            "О золотой карте в Санкт-Петербурге",
            template.render({"text": "О золотая карта в Санкт-Петербурге"})
        )

    def test_make_agree_with_number_1(self):
        template = jinja2.Template("Температура: {{temperature}} {{'градус'|make_agree_with_number(temperature)}}")
        self.assertEqual("Температура: 21 градус", template.render({"temperature": 21}))

    def test_make_agree_with_number_2(self):
        template = jinja2.Template("Температура: {{temperature}} {{'градус'|make_agree_with_number(temperature)}}")
        self.assertEqual("Температура: 25 градусов", template.render({"temperature": 25}))

    def test_comma_float_ordinar_float(self):
        template = jinja2.Template("{{'1.23'|comma_float * 2}}")
        self.assertEqual("2.46", template.render())

    def test_comma_float_comma_float(self):
        template = jinja2.Template("{{'1,23'|comma_float * 2}}")
        self.assertEqual("2.46", template.render())

    def test_careful_int_true_int(self):
        template = jinja2.Template("{{t|careful_int * 2}}")
        self.assertEqual("2", template.render({"t": 1.0}))

    def test_careful_int_false_int(self):
        template = jinja2.Template("{{t|careful_int * 2}}")
        self.assertEqual("2.46", template.render({"t": 1.23}))

    def test_tojson(self):
        template = jinja2.Template("{{t|tojson}}")
        self.assertEqual('{"word": 1, "n": null}', template.render({"t": {"word": 1, "n": None}}))

    def test_fromjson(self):
        template = jinja2.Template("{{t|fromjson}}")
        self.assertEqual("{'word': 1, 'n': None}", template.render({"t": '{"word": 1, "n": null}'}))

    def test_num2text_natural(self):
        template = jinja2.Template("{{t|num2text}} {{'градус'|make_agree_with_number(t|abs)}}")
        self.assertEqual("двадцать один градус", template.render({"t": 21}))

    def test_num2text_real(self):
        template = jinja2.Template("{{t|num2text}} {{'градус'|make_agree_with_number(t|abs)}}")
        self.assertEqual("минус семнадцать с половиной градусов", template.render({"t": -17.5}))

    def test_num2text_usual_number(self):
        self.assertEqual("одиннадцать", self.num2text(11))

    def test_num2text_big_number(self):
        self.assertEqual("сто шестьдесят семь тысяч триста сорок два", self.num2text(167342))

    def test_num2text_frac_number(self):
        self.assertEqual("двенадцать целых тридцать четыре сотых", self.num2text(12.34))

    def test_num2text_half_frac_number(self):
        self.assertEqual("семнадцать с половиной", self.num2text(17.5))

    def test_num2text_m(self):
        self.assertEqual("два", self.num2text(2))

    def test_num2text_f(self):
        self.assertEqual("две", self.num2text(2, "f"))

    def test_num2text_str(self):
        self.assertEqual("два", self.num2text("2"))

    def test_num2text_negative_real(self):
        self.assertEqual("минус две целых шестьдесят семь сотых", self.num2text(-2.67))

    def test_num2ord_gen_neut(self):
        self.assertEqual("восьмого", self.num2ord(8, 'n', 'gent'))

    def test_jinja_num2text_half_frac_number(self):
        template = jinja2.Template("{{t|num2text}} градусов")
        self.assertEqual("семнадцать с половиной градусов", template.render({"t": 17.5}))

    def test_jinja_num2ord_gen_neut(self):
        template = jinja2.Template("{{t|num2ord('n','gent')}} сентября")
        self.assertEqual("восьмого сентября", template.render({"t": 8}))

    def test_jinja_num2ord_nom_neut(self):
        template = jinja2.Template("{{t|num2ord('n')}} сентября")
        self.assertEqual("восьмое сентября", template.render({"t": 8}))

    def test_jinja_num2ord_replace_all_explicit(self):
        template = jinja2.Template("{{t|num2ord_in_text(only_explicit=True)}}")
        self.assertEqual("двадцать пятый день, двадцать шестая ночь, 27 утро",
                         template.render({"t": "25ый день, 26ая ночь, 27 утро"}))

    def test_jinja_num2ord_replace_all_not_explicit(self):
        template = jinja2.Template("{{t|num2ord_in_text(sex='n')}}")
        self.assertEqual("двадцать пятый день, двадцать шестая ночь, двадцать седьмое утро",
                         template.render({"t": "25ый день, 26ая ночь, 27 утро"}))

    def test_jinja_num2text_replace_all(self):
        template = jinja2.Template("{{t|num2text_in_text(sex='f')}}")
        self.assertEqual("три дня, две ночи",
                         template.render({"t": "3 дня, 2 ночи"}))

    def test_jinja_money2text_special_no_frac(self):
        num, cur = 11.0, 'USD'
        expected = "11 долларов США"
        self.assertEqual(expected, self.money2text(num, cur))

    def test_jinja_money2text_special_with_frac(self):
        num, cur = 11.99, 'USD'
        expected = "11 долларов США 99 центов"
        self.assertEqual(expected, self.money2text(num, cur))

    def test_jinja_money2text_regular_with_frac(self):
        num, cur = 1.99, 'TND'
        expected = "1.99 тунисских динаров"
        self.assertEqual(expected, self.money2text(num, cur))

    def test_jinja_money2text_regular_no_frac(self):
        num, cur = 1, 'CNY'
        expected = "1 юань"
        self.assertEqual(expected, self.money2text(num, cur))

    def test_ics(self):
        date = datetime.datetime.now()

        expected = "".join(map(str, [date.year, "{:02d}".format(date.month), "{:02d}".format(date.day)]))
        got = core.unified_template.jinja_filters.generate_ics(date)
        timestring = got.split()[2] #Third row contains date
        self.assertTrue(expected in timestring)

    def test_date_from_timestamp(self):
        tstamp = 0
        expected = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
        got = core.unified_template.jinja_filters.timestamp_to_datetime(tstamp)
        self.assertEqual(expected, got)

    def test_wrap_datetime(self):
        dt = [1, 1, 2019, 0, 0]
        got = core.unified_template.jinja_filters.wrap_datetime(*dt)
        expected = datetime.datetime(day=1, month=1, year=2019, hour=0, minute=0, tzinfo=datetime.timezone.utc)
        self.assertEqual(expected, got)

    def test_add_timedelta(self):
        date = datetime.datetime.now()
        day = 1
        expected = date + datetime.timedelta(days=day)
        self.assertEqual(expected, core.unified_template.jinja_filters.add_timedelta(date, days=day))

    def test_cardtype_visa(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "4716333884162944"})
        self.assertEqual("visa", result)

    def test_cardtype_master(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "5122185757184370"})
        self.assertEqual("mastercard", result)

    def test_cardtype_maestro(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "6762940925503833"})
        self.assertEqual("maestro", result)

    def test_cardtype_amex(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "373268674871118"})
        self.assertEqual("amex", result)

    def test_cardtype_mir(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "2222 2222 2222 2222 2"})
        self.assertEqual("mir", result)

    def test_cardtype_visa_masked(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "471633****2944"})
        self.assertEqual("visa", result)

    def test_cardtype_visa_empty(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": ""})
        self.assertEqual("None", result)

    def test_cardtype_visa_formatted(self):
        template = jinja2.Template("{{t|pan2service}}")
        result = template.render({"t": "4716 3311 2222 2944"})
        self.assertEqual("visa", result)

    def test_timestamp_to_date_filter(self):
        self.assertEqual(core.unified_template.jinja_filters.timestamp_to_date(1582890903), "28 февраля", "sec")
        self.assertEqual(core.unified_template.jinja_filters.timestamp_to_date(1582890903000), "28 февраля", "ms")
        self.assertEqual(core.unified_template.jinja_filters.timestamp_to_date(1582890903000000), "28 февраля", "mcs")



# coding: utf-8
from core.model.factory import build_factory
from core.model.registered import Registered
from core.monitoring.monitoring import monitoring
from core.utils.rerunable import Rerunable

db_adapters = Registered()
db_adapter_factory = build_factory(db_adapters)


class DBAdapterException(Exception):
    pass


class DBAdapter(Rerunable):
    IS_ASYNC = False

    def __init__(self, config=None):
        super(DBAdapter, self).__init__(config)
        self._client = None

    def _on_prepare(self):
        raise NotImplementedError

    def connect(self):
        raise NotImplementedError

    def _open(self, filename, *args, **kwargs):
        raise NotImplementedError

    def _save(self, id, data):
        raise NotImplementedError

    def _replace_if_equals(self, id, sample, data):
        raise NotImplementedError

    def _get(self, id):
        raise NotImplementedError

    def _list_dir(self, path):
        raise NotImplementedError

    def _glob(self, path, pattern):
        raise NotImplementedError

    def _path_exists(self, path):
        raise NotImplementedError

    def _mtime(self, path):
        raise NotImplementedError

    def open(self, filename, *args, **kwargs):
        return self._run(self._open, filename, *args, **kwargs)

    def glob(self, path, pattern):
        return self._run(self._glob, path, pattern)

    def path_exists(self, path):
        return self._run(self._path_exists, path)

    def mtime(self, path):
        return self._run(self._mtime, path)

    @monitoring.got_histogram("save_time")
    def save(self, id, data):
        return self._run(self._save, id, data)

    @monitoring.got_histogram("save_time")
    def replace_if_equals(self, id, sample, data):
        return self._run(self._replace_if_equals, id, sample, data)

    @monitoring.got_histogram("get_time")
    def get(self, id):
        return self._run(self._get, id)

    def list_dir(self, path):
        return self._run(self._list_dir, path)

    @property
    def _handled_exception(self):
        return Exception

    def _on_all_tries_fail(self):
        raise



# coding: utf-8
import unittest

from core.basic_models.operators.operators import MoreOperator, \
    LessOperator, MoreOrEqualOperator, LessOrEqualOperator, EqualOperator, \
    EndsWithOperator, Exists, StartsWithOperator, InOperator


class MoreOperatorTest(unittest.TestCase):
    def test_more_compare_sucsess(self):
        operator = MoreOperator({"amount": 1})
        assert operator.compare(3)

    def test_more_compare_fail(self):
        operator = MoreOperator({"amount": 10})
        assert not operator.compare(3)

    def test_less_compare_sucsess(self):
        operator = LessOperator({"amount": 10})
        assert operator.compare(3)

    def test_less_compare_fail(self):
        operator = LessOperator({"amount": 1})
        assert not operator.compare(3)

    def test_more_or_equal_sucsess(self):
        operator = MoreOrEqualOperator({"amount": 1})
        assert operator.compare(1)

    def test_more_or_equal_fail(self):
        operator = MoreOrEqualOperator({"amount": 3})
        assert not operator.compare(1)

    def test_less_or_equal_sucsess(self):
        operator = LessOrEqualOperator({"amount": 1})
        assert operator.compare(1)

    def test_less_or_equal_fail(self):
        operator = LessOrEqualOperator({"amount": 1})
        assert not operator.compare(3)

    def test_equal_sucsess(self):
        operator = EqualOperator({"amount": 1})
        assert operator.compare(1)

    def test_equal_fail(self):
        operator = EqualOperator({"amount": 2})
        assert not operator.compare(1)

    def test_in_success(self):
        operator = InOperator({"amount": ["val1", "val2"]})
        assert operator.compare("val1")

    def test_in_fail(self):
        operator = InOperator({"amount": ["val1", "val2"]})
        self.assertFalse(operator.compare("another_value"))

    def test_endswith_success_with_string(self):
        operator = EndsWithOperator({'amount': 'abcd'})
        assert operator.compare('abbbabcd')

    def test_endswith_success_with_digits(self):
        operator = EndsWithOperator({'amount': 7890})
        assert operator.compare('1234567890')

    def test_endswith_success_with_digits_2(self):
        operator = EndsWithOperator({'amount': '7890'})
        assert operator.compare(1234567890)

    def test_endswith_fail_with_string(self):
        operator = EndsWithOperator({'amount': 'abcd'})
        assert not operator.compare('abbbabcda')

    def test_endswith_fail_with_digits(self):
        operator = EndsWithOperator({'amount': 78901})
        assert not operator.compare('1234567890')

    def test_endswith_fail_with_digits_2(self):
        operator = EndsWithOperator({'amount': '78901'})
        assert not operator.compare(1234567890)

    def test_startswith_success_with_string(self):
        operator = StartsWithOperator({'amount': 'abbb'})
        assert operator.compare('abbbabcd')

    def test_startswith_success_with_digits(self):
        operator = StartsWithOperator({'amount': 1234})
        assert operator.compare('1234567890')

    def test_startswith_success_with_digits_2(self):
        operator = StartsWithOperator({'amount': '1234'})
        assert operator.compare(1234567890)

    def test_startswith_fail_with_string(self):
        operator = StartsWithOperator({'amount': 'abbc'})
        assert not operator.compare('abbbabcda')

    def test_startswith_fail_with_digits(self):
        operator = StartsWithOperator({'amount': 1233})
        assert not operator.compare('1234567890')

    def test_startswith_fail_with_digits_2(self):
        operator = StartsWithOperator({'amount': '1233'})
        assert not operator.compare(1234567890)

    def test_exists_success(self):
        operator = Exists({})
        self.assertTrue(operator.compare('value'))

    def test_exists_fail(self):
        operator = Exists({})
        self.assertFalse(operator.compare(None))



# coding: utf-8
import logging
import os
import time
from typing import Union

from confluent_kafka import Producer

import core.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.monitoring.monitoring import monitoring
from core.mq.kafka.base_kafka_publisher import BaseKafkaPublisher


class KafkaPublisher(BaseKafkaPublisher):
    def __init__(self, config):
        self._config = config["publisher"]
        conf = self._config["conf"]
        conf["error_cb"] = self._error_callback
        conf["on_delivery"] = self._delivery_callback
        internal_log_path = self._config.get("internal_log_path")
        if internal_log_path:
            debug_logger = logging.getLogger("debug_publisher")
            timestamp = time.strftime("_%d%m%Y_")
            debug_logger.addHandler(logging.FileHandler(
                "{}/kafka_publisher_debug{}{}.log".format(internal_log_path, timestamp, os.getpid())
            ))
            conf["logger"] = debug_logger
        self._producer = Producer(**conf)

    def send(self, value: Union[str, bytes], key=None, topic_key=None, headers=None):
        try:
            topic = self._config["topic"]
            if topic_key is not None:
                topic = topic[topic_key]
            producer_params = dict()
            if key is not None:
                producer_params["key"] = key
            self._producer.produce(topic=topic, value=value, headers=headers or [], **producer_params)
        except BufferError as e:
            params = {
                "queue_amount": len(self._producer),
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("KafkaProducer: Local producer queue is full (%(queue_amount)s messages awaiting delivery):"
                       " try again\n", params=params, level="ERROR")
            monitoring.got_counter("kafka_producer_exception")
        self._poll()

    def send_to_topic(self, value, key=None, topic=None, headers=None):
        try:
            if topic is None:
                params = {
                    "message": str(value),
                    log_const.KEY_NAME: log_const.EXCEPTION_VALUE
                }
                log("KafkaProducer: Failed sending message %{message}s. Topic is not defined", params=params,
                              level="ERROR")
            producer_params = dict()
            if key is not None:
                producer_params["key"] = key
            self._producer.produce(topic=topic, value=value, headers=headers or [], **producer_params)
        except BufferError as e:
            params = {
                "queue_amount": len(self._producer),
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("KafkaProducer: Local producer queue is full (%(queue_amount)s messages awaiting delivery):"
                       " try again\n", params=params, level="ERROR")
            monitoring.got_counter("kafka_producer_exception")
        self._poll()

    def _poll(self):
        poll_timeout = self._config.get("poll_timeout", 0)
        if poll_timeout > 0:
            while True:
                result = self._producer.poll(poll_timeout)
                if not result:
                    return

    def _error_callback(self, err):
        params = {
            "error": str(err),
            log_const.KEY_NAME: log_const.EXCEPTION_VALUE
        }
        log("KafkaProducer: Error: %(error)s", params=params, level="ERROR")
        monitoring.got_counter("kafka_producer_exception")

    def _delivery_callback(self, err, msg):
        if err:
            message_text = msg.value()
            try:
                message_text = message_text.decode("utf-8")
                log("KafkaProducer: Message %(message)s send failed: %(error)s",
                              params={
                                  "message": str(message_text),
                                  "error": str(err),
                                  log_const.KEY_NAME: log_const.EXCEPTION_VALUE},
                              level="ERROR")
            except UnicodeDecodeError:
                log("KafkaProducer: %(text)s: %(error)s",
                              params={"text": f"Can't decode: {str(message_text)}",
                                      "error": err,
                                      log_const.KEY_NAME: log_const.EXCEPTION_VALUE},
                              level="ERROR",
                              exc_info=True)
            monitoring.got_counter("kafka_producer_exception")

    def close(self):
        self._producer.flush(self._config["flush_timeout"])



# coding: utf-8
import time


class LastField:
    def __init__(self, items):
        items = items or {}
        self.value = items.get("value")
        self.remove_time = items.get("remove_time")

    def set_remove_time(self, lifetime):
        if lifetime:
            self.remove_time = int(time.time()) + lifetime

    @property
    def raw(self):
        if self.value or self.remove_time:
            return {"value": self.value,
                    "remove_time": self.remove_time}



import io
import time
from prometheus_client import CONTENT_TYPE_LATEST
from prometheus_client.twisted import MetricsResource
from twisted.web.resource import Resource
from core.monitoring.twisted_server import TwistedServer
from core.utils.memstats import get_meminfo, show_growth, show_most_common_types, get_leaking_objects


def add_headers(request, response_text):
    request.responseHeaders.addRawHeader(b"content-type", CONTENT_TYPE_LATEST.encode())
    request.responseHeaders.addRawHeader(b"content-length", str(len(response_text)).encode())


class RootResource(Resource):

    def __init__(self, debug=False):
        super(RootResource, self).__init__()
        self.putChild(b'health', HealthcheckResource())
        self.putChild(b'metrics', MetricsResource())

        if debug:
            self.putChild(b'meminfo', MemInfoResource())
            self.putChild(b'objgrowth', ObjGrowthResource())
            self.putChild(b'objtypes', ObjTypesResource())
            self.putChild(b'objleak', ObjLeakResource())

    def getChild(self, name, request):
        if name:
            return Resource.getChild(self, name, request)
        return self

    def render_GET(self, request):
        response = ""
        add_headers(request, response)
        return response.encode()


class HealthcheckResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        response = "ok"
        add_headers(request, response)
        return response.encode()


class MemInfoResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        response = get_meminfo()
        add_headers(request, response)
        return response.encode()


class ObjGrowthResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        with io.StringIO() as tempio:
            show_growth(file=tempio)
            response = tempio.getvalue()

        add_headers(request, response)
        return response.encode()


class ObjTypesResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        with io.StringIO() as tempio:
            show_most_common_types(file=tempio, limit=20)
            response = tempio.getvalue()
        add_headers(request, response)
        return response.encode()


class ObjLeakResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        with io.StringIO() as tempio:
            get_leaking_objects(file=tempio, limit=5)
            response = tempio.getvalue()
        add_headers(request, response)
        return response.encode()


if __name__ == "__main__":
    t = TwistedServer(1111, "localhost", RootResource, debug=True)
    while 1:
        t.iterate()
        time.sleep(1)



# coding: utf-8
import unittest

from core.message import device


class MessageDeviceTest1(unittest.TestCase):
    def test_device1(self):
        obj = device.Device(dict())
        self.assertTrue(obj.value == {})
        self.assertTrue(obj.platform_type == "")
        self.assertTrue(obj.platform_version == "")
        self.assertTrue(obj.surface == "")
        self.assertTrue(obj.surface_version == "")
        self.assertTrue(obj.features == {})
        self.assertTrue(obj.capabilities == {})
        self.assertTrue(obj.additional_info == {})

    def test_device2(self):
        test_dict = {"platformType": "1", "platformVersion": "1", "surface": "1", "surfaceVersion": "1",
                     "features": {1: 1}, "capabilities": {1: 1}, "additionalInfo": {1: 1}}
        obj = device.Device(test_dict)
        self.assertTrue(obj.value == test_dict)
        self.assertTrue(obj.platform_type == "1")
        self.assertTrue(obj.platform_version == "1")
        self.assertTrue(obj.surface == "1")
        self.assertTrue(obj.surface_version == "1")
        self.assertTrue(obj.features == {1: 1})
        self.assertTrue(obj.capabilities == {1: 1})
        self.assertTrue(obj.additional_info == {1: 1})



from core.descriptions.descriptions_items import DescriptionsItems


class LimitedQueuedHashableObjectsDescription:
    DEFAULT_MAX_LEN = 10

    def __init__(self, items, id=None):
        self.id = id
        self.items = items or {}
        self.max_len = self.items.get("max_len", self.DEFAULT_MAX_LEN)


class LimitedQueuedHashableObjectsDescriptionsItems(DescriptionsItems):
    def __init__(self, items):
        super(LimitedQueuedHashableObjectsDescriptionsItems, self).__init__(LimitedQueuedHashableObjectsDescription, items)



import inspect
import pickle
import sys
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Dict, Optional, Union, List

import numpy as np
from timeout_decorator import timeout_decorator

import core.basic_models.classifiers.classifiers_constants as cls_const
from core.basic_models.classifiers.vectorizer_models import vectorizers
from core.model.factory import build_factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.utils.exception_handlers import exc_handler

classifiers = Registered()

classifier_factory = build_factory(classifiers)


class Classifier(ABC):
    """Базовый класс для сущности Классификатор."""

    CLASSIFIER_TYPE = None

    def __init__(self, settings: Dict[str, Any], id: Optional[str] = None) -> None:
        self.id = id
        self.settings = settings if settings else {}
        self.version = settings.get("version", -1)
        self.threshold = self.settings.get("threshold", 0)
        self.intents = self.settings.get("intents", [])
        self.class_other = cls_const.OTHER_KEY
        self._check_classifier_type(settings["type"])

    def _answer_template(self, intent: str, score: float, is_other: bool) -> Dict[str, Union[str, float, bool]]:
        # Любой классификатор должен возвращать отсортированный список наиболее вероятных вариантов из заданного
        # множества, прошедших определенный порог уверенности. Каждый вариант из списка должен соответвовать общему
        # шаблону: answer=классу, score=величине уверенности в ответе, other=булево значение (принадлежность к other).
        return {cls_const.ANSWER_KEY: intent, cls_const.SCORE_KEY: score, self.class_other: is_other}

    def _check_classifier_type(self, classifier_type: str) -> None:
        if classifier_type != self.CLASSIFIER_TYPE:
            raise Exception(f"Inappropriate classifier type: {classifier_type}, it should be {self.CLASSIFIER_TYPE}")

    @abstractmethod
    def find_best_answer(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            mask: Optional[Dict[str, bool]] = None,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Union[str, float, bool]]]:
        # Формируется отсортированный список наиболее вероятных вариантов
        raise NotImplementedError

    @abstractmethod
    def initial_launch(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> Union[List[Dict[str, Union[str, float, bool]]], None]:
        # Первоначальный запуск модели классификатора
        raise NotImplementedError


class SkipClassifier(Classifier):
    """
    Классификатор, который не делает (пропускает) сам процесс классификации.
    Используется, когда необходимо по формату указать классификатор, но использовать конкретное значение-результат.
    """

    CLASSIFIER_TYPE = "skip"

    def __init__(self, settings: Dict[str, Any], id: Optional[str] = None) -> None:
        super(SkipClassifier, self).__init__(settings, id)
        self.intents = self.settings["intents"]

    def find_best_answer(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            mask: Optional[Dict[str, bool]] = None,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Union[str, float, bool]]]:
        return [self._answer_template(intent, 0, False) for intent in self.intents]

    def initial_launch(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> Union[List[Dict[str, Union[str, float, bool]]], None]:
        pass

    @staticmethod
    def get_nothing() -> Dict[str, Any]:
        return {"type": "skip", "intents": []}


class ExternalClassifier(Classifier):
    """Внешний классификатор.
    Выполняет некую функцию обёртки для вызова реализованных классов классификаторов по имени.
    """

    # Дефолтное значение таймаута, время за которое должен прийти ответ от внешнего классификатора
    BLOCKING_TIMEOUT = cls_const.EXTERNAL_CLASSIFIER_BLOCKING_TIMEOUT
    CLASSIFIER_TYPE = "external"

    def __init__(self, settings: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ExternalClassifier, self).__init__(settings, id)
        self._classifier_key = settings["classifier"]
        self._timeout_wrap = timeout_decorator.timeout(self.settings.get("timeout") or self.BLOCKING_TIMEOUT)

    @exc_handler(on_error_obj_method_name="on_timeout_error", handled_exceptions=(timeout_decorator.TimeoutError,))
    def find_best_answer(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            mask: Optional[Dict[str, bool]] = None,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Union[str, float, bool]]]:
        classifier = scenario_classifiers[self._classifier_key]
        return self._timeout_wrap(classifier.find_best_answer)(text_preprocessing_result, mask, scenario_classifiers)

    @staticmethod
    def on_timeout_error(*args, **kwarg):
        return list()

    def initial_launch(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ):
        classifier = scenario_classifiers[self._classifier_key]
        return classifier.initial_launch(text_preprocessing_result, scenario_classifiers)


class ExtendedClassifier(Classifier):
    """Класс не является самостоятельным типом классификатора. Расширяет функционал базового класса."""

    def __init__(self, settings: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ExtendedClassifier, self).__init__(settings, id)
        self.intents = self.settings["intents"]
        self._path = self.settings["path"]
        self._classifier = self.settings.get("classifier")
        # Способ векторизации для реплик указывается в конфигурации классификатора
        self._vectorizer = self.settings.get("vectorizer")

    def set_classifier(self, clsf: Classifier) -> None:
        self._classifier = clsf

    @cached_property
    def classifier(self) -> Classifier:
        return self._classifier

    def find_best_answer(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            mask: Optional[Dict[str, bool]] = None,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Union[str, float, bool]]]:
        vector = vectorizers[self._vectorizer].vectorize(text_preprocessing_result) if self._vectorizer else np.array([])
        weights = sorted(self._get_weights(text_preprocessing_result, vector).items(), key=lambda x: x[1], reverse=True)
        answers = []
        for weight in weights:
            if weight[0] < len(self.intents):
                cls_name = self.intents[weight[0]]
                cls_prob = weight[1]
                answers.append(self._answer_template(cls_name, cls_prob, cls_name == self.class_other))
        return answers

    def _get_weights(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            vector: Optional[np.ndarray] = np.array([]),
            numb: int = 3
    ):
        weights = self._prediction(text_preprocessing_result, vector)
        tuple_weights = sorted(
            {i: weight for i, weight in enumerate(weights) if weight >= self.threshold}.items(),
            key=lambda x: x[1],
            reverse=True
        )
        tuple_weights = tuple_weights[:numb]  # берем numb наибольших значений весов
        return dict(tuple_weights)

    @abstractmethod
    def _prediction(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            vector: Optional[np.ndarray] = np.array([])
    ) -> List[Any]:
        raise NotImplementedError

    def initial_launch(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            scenario_classifiers: Optional[Dict[str, Any]] = None
    ) -> Union[List[Dict[str, Union[str, float, bool]]], None]:
        return self.find_best_answer(text_preprocessing_result, None, classifiers)


class SciKitClassifier(ExtendedClassifier):
    """Класс для загрузки и инфера моделей обученных с помощью библиотеки sklearn и имеющих тип scikit.
    У сохраненного класса обученной модели предполагается обязательное наличие метода predict_proba.
    """

    CLASSIFIER_TYPE = "scikit"

    def __init__(self, settings: Dict[str, Any], id: Optional[str] = None) -> None:
        super(SciKitClassifier, self).__init__(settings, id)

    @staticmethod
    def prepared(text_preprocessing_result: BaseTextPreprocessingResult):
        return pickle.dumps(text_preprocessing_result.tokenized_elements_list)

    def _prediction(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            vector: Optional[np.ndarray] = np.array([])
    ) -> List[Any]:
        if vector.size != 0:
            prediction_result = self.classifier.predict_proba(
                self.prepared(text_preprocessing_result), vector)[0].tolist()
        else:
            prediction_result = self.classifier.predict_proba(self.prepared(text_preprocessing_result))[0].tolist()
        return prediction_result


# Реализованные на данный момент типы классификаторов
SUPPORTED_CLASSIFIERS_TYPES = frozenset([
    class_tuple[1].CLASSIFIER_TYPE
    for class_tuple in inspect.getmembers(sys.modules[__name__], inspect.isclass)
    if hasattr(class_tuple[1], "CLASSIFIER_TYPE") and class_tuple[1].CLASSIFIER_TYPE
])



import core.names.field as core_field

CALLBACK_ID_HEADER = "app_callback_id"
LINK_BEHAVIOR_FLAG = "link_previous_behavior"
KAFKA = "kafka"
KAFKA_REPLY_TOPIC = "kafka_replyTopic"

ORIGINAL_INTENT = "original_intent"
NEW_SESSION = "new_session"

SAVED_BEHAVIOR_PARAMS_FIELDS = {core_field.INTENT, core_field.INTENT_META,
                                ORIGINAL_INTENT, core_field.APP_INFO, core_field.PROJECT_NAME, NEW_SESSION,
                                core_field.APPLICATION_ID, core_field.APP_VERSION_ID}



# coding: utf-8
import unittest
from unittest.mock import Mock

from smart_kit.request import kafka_request


class RequestTest1(unittest.TestCase):
    def setUp(self):
        self.test_items1 = {"topic_key": "12345", "kafka_replyTopic": "any theme", "timeout": 10,
                           "app_callback_id": 22, "kafka_key": "54321"}  # пусть
        self.test_items2 = {"topic_key": "12345", "timeout": 10}  # пусть
        self.test_items3 = {"topic_key": "12345", "kafka_replyTopic": "any theme", "timeout": 10}  # пусть
        self.test_items4 = {"topic_key": "12345", "kafka_replyTopic": "any theme", "timeout": 10, "topic": "topic123"}  # пусть
        self.test_source_mq_message = Mock('source_mq_message')
        self.test_source_mq_message.headers = lambda: [("any header 1", 1)]

    def test_smart_kafka_request_init(self):
        obj1 = kafka_request.SmartKitKafkaRequest(self.test_items1)
        self.assertTrue(obj1.KAFKA_REPLY_TOPIC == "kafka_replyTopic")
        self.assertTrue(obj1.KAFKA_KEY == "kafka_key")
        self.assertTrue(obj1.topic_key == "12345")
        with self.assertRaises(AttributeError):
            obj2 = kafka_request.SmartKitKafkaRequest("")
        self.assertTrue(obj1._callback_id == 22)
        self.assertTrue(obj1._kafka_replyTopic == "any theme")

    def test_smart_kafka_request_topic(self):
        obj1 = kafka_request.SmartKitKafkaRequest(self.test_items4)
        self.assertTrue(obj1.topic_key == "12345")
        self.assertTrue(obj1.topic == "topic123")

    def test_smart_kafka_request_callback_id_header_name(self):
        obj1 = kafka_request.SmartKitKafkaRequest(self.test_items1)
        self.assertTrue(obj1._callback_id_header_name == "app_callback_id")

    def test_smart_kafka_request_get_new_headers(self):
        obj1 = kafka_request.SmartKitKafkaRequest(self.test_items1)
        obj2 = kafka_request.SmartKitKafkaRequest(self.test_items2)
        obj3 = kafka_request.SmartKitKafkaRequest(self.test_items3)
        self.assertTrue(obj1._get_new_headers(self.test_source_mq_message) == [('any header 1', 1), (
                                                                               'app_callback_id', b'22'),
                                                                               ('kafka_replyTopic', b'any theme')])
        self.assertTrue(obj2._get_new_headers(self.test_source_mq_message) == [('any header 1', 1)])
        self.assertTrue(obj3._get_new_headers(self.test_source_mq_message) == [('any header 1', 1),
                                                                               ('kafka_replyTopic', b'any theme')])

    def test_smart_kafka_request_str(self):
        obj1 = kafka_request.SmartKitKafkaRequest(self.test_items1)
        obj2 = kafka_request.SmartKitKafkaRequest(self.test_items2)
        self.assertTrue(obj1.__str__() == "KafkaRequest: kafka_key=54321")
        self.assertTrue(obj2.__str__() == "KafkaRequest: kafka_key=None")



import requests
import os
from tqdm import tqdm
from typing import List, Sequence

from smart_kit.text_preprocessing.base_text_normalizer import BaseTextNormalizer
from smart_kit.utils.cache import Cache, ItemExpired


def words_tokenized_set(text):
    return set(text.split(' '))


class NormalizationError(Exception):
    pass


class HttpTextNormalizer(BaseTextNormalizer):

    def __init__(self, url: str, batch_size: int = 128, timeout=10, verbose=False, cache_lifetime=0):
        url = url.rstrip("/") + "/"
        self._preprocess_url = url + HttpTextNormalizer.PREPROCESS_METHOD
        self._classify_url = url + HttpTextNormalizer.CLASSIFY_METHOD
        self._normalize_url = url + HttpTextNormalizer.NORMALIZE_METHOD
        self._batch_size = batch_size
        self._timeout = timeout
        self._verbose = verbose
        self._tqdm_func = tqdm if verbose else (lambda x: x)
        self.set_normalize_mode()

        self.cache = self.CACHE(cache_lifetime)

    @classmethod
    def with_cache(cls, app_config, **kwargs):
        cache_impl = app_config.NORMALIZATION_CACHE
        cls.CACHE = cache_impl
        params = {"url": app_config.NORMALIZER_ADDRESS, "cache_lifetime": app_config.NORMALIZATION_CACHE_TTL}
        params.update(kwargs)
        inst = cls(**params)
        try:
            inst.cache.load(os.path.join(app_config.REFERENCES_PATH, ".normalization_cache.json"))
        except FileNotFoundError:
            pass
        return inst

    def print_current_state(self):
        if self._url == self._preprocess_url:
            print("Выбран режим PREPROCESS_TEXT")
        elif self._url == self._classify_url:
            print("Выбран режим CLASSIFY_TEXT")
        elif self._url == self._normalize_url:
            print("Выбран режим NORMALIZE (только предобработка без аннотаторов)")
        else:
            print("Выбран неизвестный режим")

    def set_preprocess_mode(self):
        self._url = self._preprocess_url
        if self._verbose:
            self.print_current_state()

    def set_classify_mode(self):
        self._url = self._classify_url
        if self._verbose:
            self.print_current_state()

    def set_normalize_mode(self):
        self._url = self._normalize_url
        if self._verbose:
            self.print_current_state()

    def _get_batch(self, sequence: Sequence, n=1) -> Sequence:
        max_ndx = len(sequence)
        for index in self._tqdm_func(range(0, max_ndx, n)):
            yield sequence[index: min(index + n, max_ndx)]

    def _normalize_batch(self, batch: Sequence) -> List:
        response = requests.post(self._url, json=[{HttpTextNormalizer.TEXT_PARAM_NAME: el} for el in batch],
                                 timeout=self._timeout)
        response.raise_for_status()
        result = [item["message"] for item in response.json()]
        for item in result:
            self.cache[item["original_text"]] = item
        return result

    def normalize_sequence(self, texts: Sequence, batch_size=None) -> List:
        normalized_texts = []
        batch_generator = self._get_batch(texts, batch_size or self._batch_size)
        for batch in batch_generator:
            normalized_texts.extend(self._normalize_batch(batch))
        return normalized_texts

    def __call__(self, text: str):
        try:
            return self.cache[text]
        except (KeyError, ItemExpired):
            pass

        batch = self._normalize_batch([text])
        if len(batch) == 0:
            raise NormalizationError("Empty Message Received")

        return batch[0]

    def load_everything(self) -> None:
        pass



from unittest.mock import Mock, MagicMock


class PicklableMock(Mock):
    def __reduce__(self):
        return Mock, ()


class PicklableMagicMock(MagicMock):
    def __reduce__(self):
        return MagicMock, ()



# coding: utf-8
import socket
import unittest
from collections import OrderedDict
from collections import namedtuple
from unittest.mock import Mock

import scenarios.behaviors.behaviors
from smart_kit.utils.picklable_mock import PicklableMock


class BehaviorsTest(unittest.TestCase):
    def setUp(self):
        self.user = PicklableMock()
        self.user.settings = PicklableMock()
        self.user.settings.app_name = "app_name"
        self.user.local_vars.values = {"test_local_var_key": "test_local_var_value"}
        self.description = PicklableMock()
        self.description.timeout = Mock(return_value=10)
        self.success_action = PicklableMock()
        self.success_action.run = PicklableMock()
        self.fail_action = PicklableMock()
        self.timeout_action = PicklableMock()

        self.description.success_action = self.success_action
        self.description.fail_action = self.fail_action
        self.description.timeout_action = self.timeout_action
        self.descriptions = {"test": self.description}
        self._callback = namedtuple('Callback', 'behavior_id expire_time scenario_id')

    def test_success(self):
        callback_id = "123"
        behavior_id = "test"
        item = {"behavior_id": behavior_id, "expire_time": 2554416000, "scenario_id": None,
                "text_preprocessing_result": {}}
        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        behaviors.success(callback_id)
        # self.success_action.run.assert_called_once_with(self.user, TextPreprocessingResult({}))
        self.success_action.run.assert_called_once()
        self.assertDictEqual(behaviors.raw, {})

    def test_success_2(self):
        callback_id = "123"
        items = {}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        behaviors.success(callback_id)
        self.success_action.run.assert_not_called()

    def test_fail(self):
        callback_id = "123"
        behavior_id = "test"
        item = {"behavior_id": behavior_id, "expire_time": 2554416000, "scenario_id": None}
        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        behaviors.fail(callback_id)
        self.fail_action.run.assert_called_once()
        self.assertDictEqual(behaviors.raw, {})

    def test_timeout(self):
        callback_id = "123"
        behavior_id = "test"
        item = {"behavior_id": behavior_id, "expire_time": 2554416000, "scenario_id": None}
        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        behaviors.timeout(callback_id)
        self.timeout_action.run.assert_called_once()
        self.assertDictEqual(behaviors.raw, {})

    def test_expire(self):
        callback_id = "123"
        behavior_id = "test"
        item = {"behavior_id": behavior_id, "expire_time": 1548079039, "scenario_id": None}
        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        behaviors.expire()
        self.assertDictEqual(behaviors.raw, {})

    @unittest.mock.patch.object(scenarios.behaviors.behaviors, "time", return_value=9999999999)
    def test_add_1(self, time):
        items = {}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        callback_id = "123"
        behavior_id = "test"
        text_preprocessing_result = {}
        behaviors.add(callback_id, behavior_id)
        _time = int(time()) + self.description.timeout(None) + scenarios.behaviors.behaviors.Behaviors.EXPIRATION_DELAY

        exp = OrderedDict(behavior_id=behavior_id, expire_time=_time, scenario_id=None,
                          text_preprocessing_result=text_preprocessing_result,
                          action_params={'local_vars': {'test_local_var_key': 'test_local_var_value'}},
                          hostname=socket.gethostname())
        self.assertDictEqual(behaviors.raw, {callback_id: exp})

    @unittest.mock.patch.object(scenarios.behaviors.behaviors, "time", return_value=9999999999)
    def test_add_2(self, time):
        items = {}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        callback_id = "123"
        behavior_id = "test"
        scenario_id = "test_scen"
        text_preprocessing_result = {"test_scen": 1}

        behaviors.add(callback_id, behavior_id, scenario_id, text_preprocessing_result)
        _time = int(time()) + self.description.timeout(None) + scenarios.behaviors.behaviors.Behaviors.EXPIRATION_DELAY
        exp = OrderedDict(behavior_id=behavior_id, expire_time=_time, scenario_id=scenario_id,
                          text_preprocessing_result=text_preprocessing_result,
                          action_params={'local_vars': {'test_local_var_key': 'test_local_var_value'}},
                          hostname=socket.gethostname())
        self.assertDictEqual(behaviors.raw, {callback_id: exp})

    def test_check_1(self):
        callback_id = "123"
        behavior_id = "test"
        item = {"behavior_id": behavior_id, "expire_time": 1548079039, "scenario_id": None}
        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        self.assertTrue(behaviors.check_got_saved_id(behavior_id))

    def test_check_2(self):
        callback_id = "123"
        behavior_id = "test"
        item = {"behavior_id": behavior_id, "expire_time": 1548079039, "scenario_id": None}
        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        with self.assertRaises(KeyError):
            behaviors.check_got_saved_id("test2")

    def test_check_misstate_1(self):
        callback_id = "123"
        behavior_id = "test"
        scenario_id = "test_scen"
        item = {"behavior_id": behavior_id, "expire_time": 1548079039, "scenario_id": scenario_id}
        items = {str(callback_id): item}
        self.user.last_scenarios = PicklableMock()
        self.user.last_scenarios.last_scenario_name = "test_scen2"
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        self.assertTrue(behaviors.check_misstate(callback_id))

    def test_check_misstate_2(self):
        callback_id = "123"
        behavior_id = "test"
        scenario_id = "test_scen"
        item = {"behavior_id": behavior_id, "expire_time": 1548079039, "scenario_id": scenario_id}
        items = {str(callback_id): item}
        self.user.last_scenarios = PicklableMock()
        self.user.last_scenarios.last_scenario_name = "test_scen"
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        self.assertFalse(behaviors.check_misstate(callback_id))

    def test_raw(self):
        callback_id = "123"
        behavior_id = "test"
        scenario_id = "test_scen"
        text_preprocessing_result = {1: 2}

        item = {"behavior_id": behavior_id, "expire_time": 1548079039, "scenario_id": scenario_id,
                "text_preprocessing_result": text_preprocessing_result}

        items = {str(callback_id): item}
        behaviors = scenarios.behaviors.behaviors.Behaviors(items, self.descriptions, self.user)
        behaviors.initialize()
        expected = OrderedDict(behavior_id=behavior_id, expire_time=1548079039, scenario_id=scenario_id,
                               text_preprocessing_result=text_preprocessing_result, action_params={},
                               hostname=None)
        self.assertEqual(behaviors.raw, {callback_id: expected})



# coding: utf-8

from core.repositories.base_repository import BaseRepository


class ItemsRepository(BaseRepository):
    def __init__(self, *args, **kwargs):
        super(ItemsRepository, self).__init__(*args, **kwargs)
        self.data = dict()

    @BaseRepository.data.setter
    def data(self, value):
        if value is None:
            self._data = dict()
        else:
            self._data = value

    def load(self):
        super(ItemsRepository, self).load()

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)



# coding: utf-8
import os
import shutil
import unittest

from smart_kit.management import smart_kit_manager


class ManagementTest2(unittest.TestCase):
    def setUp(self):
        self.test_folder = "any_dir_name"

    def test_smart_kit_manager(self):
        obj1 = smart_kit_manager.SmartKitManager()
        with self.assertRaises(KeyError):
            obj1.default((1, 2, 3))

    def test_version_command(self):
        obj1 = smart_kit_manager.AppCommand()
        with self.assertRaises(NotImplementedError):
            obj1.execute(*(1, 2, 3), **{"d": 4})

    def test_create_app_command_init(self):
        obj1 = smart_kit_manager.CreateAppCommand()
        self.assertTrue(obj1.TEMPLATE_FOLDER == "template")
        self.assertTrue(obj1.extensions == (".py", ".yml"))
        self.assertTrue(hasattr(obj1, 'doc'))


class ManagementTest3(unittest.TestCase):
    def setUp(self):
        self.test_folder = "any_dir_name"

    def tearDown(self):
        shutil.rmtree(os.path.join(os.getcwd(), self.test_folder))

    def test_create_app_command_execute(self):
        obj1 = smart_kit_manager.CreateAppCommand()
        # использует try except - хотя в принципе достаточно проверить что app_name существует
        with self.assertRaises(FileExistsError):
            obj1.execute("")
        # рабочий вариант
        obj1.execute(self.test_folder)
        self.assertTrue(os.path.exists(os.path.join(os.getcwd(), self.test_folder)))
        # проверяем наличие файлов
        self.assertTrue(os.path.exists(os.path.join(os.getcwd(), self.test_folder, "manage.py")))
        self.assertTrue(os.path.exists(os.path.join(os.getcwd(), self.test_folder, "app", "user", "user.py")))
        self.assertTrue(os.path.exists(os.path.join(os.getcwd(), self.test_folder, "static", "references", "forms",
                                                    "hello_form.json")))



from core.descriptions.descriptions_items import DescriptionsItems
from core.request.base_request import request_factory


class ExternalRequests(DescriptionsItems):
    def __init__(self, items):
        super(ExternalRequests, self).__init__(request_factory, items)



# coding: utf-8
import unittest

from core.model.queued_objects.limited_queued_hashable_objects import LimitedQueuedHashableObjects
from core.model.queued_objects.limited_queued_hashable_objects_description import \
    LimitedQueuedHashableObjectsDescription


class LimitedQueuedHashableObjectsTest(unittest.TestCase):
    def setUp(self):
        self.descr = LimitedQueuedHashableObjectsDescription(None)

    def test_raw(self):
        last_messages_id = LimitedQueuedHashableObjects([1, 2, 3], self.descr)
        self.assertListEqual(last_messages_id.raw, [1, 2, 3])

    def test_add_true(self):
        last_messages_id = LimitedQueuedHashableObjects([1, 2, 3], self.descr)
        last_messages_id.check(4)
        self.assertListEqual(last_messages_id.raw, [1, 2, 3, 4])

    def test_add_false(self):
        ids = list(range(0, self.descr.max_len))
        last_messages_id = LimitedQueuedHashableObjects(ids, self.descr)
        new_id = self.descr.max_len + 1
        last_messages_id.check(new_id)
        new_ids = ids
        new_ids.pop(0)
        new_ids.append(new_id)
        self.assertListEqual(last_messages_id.raw, new_ids)

    def test_check_true(self):
        last_messages_id = LimitedQueuedHashableObjects([1, 2, 3], self.descr)
        result = last_messages_id.check(1)
        self.assertTrue(result)

    def test_check_false(self):
        last_messages_id = LimitedQueuedHashableObjects([1, 2, 3], self.descr)
        result = last_messages_id.check(4)
        self.assertFalse(result)


class LimitedQueuedHashableObjectsDescriptionTest(unittest.TestCase):
    DEFAULT_MAX_LEN = 10

    def setUp(self):
        self.expected_items = {"max_len": 2}
        self.limited_queued_hashable_objects_description = LimitedQueuedHashableObjectsDescription({"max_len": 2})

    def test_init(self):
        self.assertEqual(self.limited_queued_hashable_objects_description.items, self.expected_items)
        self.assertEqual(self.limited_queued_hashable_objects_description.max_len, 2)

    def test_default_max_len(self):
        limited_queued_hashable_objects_description = LimitedQueuedHashableObjectsDescription({})
        self.assertEqual(limited_queued_hashable_objects_description.max_len,
                         LimitedQueuedHashableObjectsDescription.DEFAULT_MAX_LEN)



"""Константы классификаторов."""


# Константы любого классификатора, формирующие шаблон ответа
SCORE_KEY = "score"
ANSWER_KEY = "answer"
OTHER_KEY = "other"

# Дополнительные константы для внешних "external" классификаторов
EXTERNAL_CLASSIFIER_BLOCKING_TIMEOUT = 0.2  # Время (в секундах) за которое классификатор должен успеть ответить

# Константы классификатора Intent Recognizer
INTENT_RECOGNIZER_ANSWER_DISTANCE_KEY = "answer_distance"
INTENT_RECOGNIZER_ANSWER_KEY = "answer"
INTENT_RECOGNIZER_OTHER_KEY = "other"

# Параметры, наличие которых обязательно в конфиге любого классификатора
REQUIRED_CONFIG_PARAMS = frozenset(["type"])
